/***************************************************************************
 *   Copyright (C) 2006 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "disassembler.h"

#include "devices/pic/pic/pic_memory.h"
#include "coff.h"

QString GPUtils::toDeviceName(const QString &device)
{
  if ( device.startsWith("PS") ) return device.lower();
  return "p" + device.lower();
}

QString GPUtils::generateDeviceConfig(const Pic::Memory &memory, bool &ok)
{
  const Pic::Data &data = memory.device();
  const Pic::Config &config = data.config();
  ok = true;
  QString s;
  for (uint i=0; i<data.nbWords(Pic::MemoryConfig); i++) {
    uint v = memory.normalizedWord(Pic::MemoryConfig, i);
    const Pic::Config::Word &cword = config._words[i];
    if ( cword.masks.count()==0 ) continue;
    s += "  __CONFIG ";
    QString cn = config.configName(i, Pic::Config::ConfigNameAsmMacro);
    if ( !cn.isEmpty() ) s += cn + ", ";
    QStringList cnames;
    for (uint k=0; k<uint(cword.masks.count()); k++) {
      const Pic::Config::Mask &cmask = cword.masks[k];
      if ( cmask.name=="DEBUG" ) continue;
      if ( (cmask.value & cword.pmask)==cmask.value ) continue; // protected bits
      for (int l=cmask.values.count()-1; l>=0; l--) {
        if ( (v & cmask.values[l].value)==cmask.values[l].value ) {
          if ( cmask.values[l].cname.isEmpty() ) ok = false;
          else cnames += cmask.values[l].cname;
          break;
        }
      }
    }
    s += cnames.join(" & ");
    s += "\n";
  }
  return s;
}

QString GPUtils::disassemble(const Pic::Memory &memory)
{
  const Pic::Data &data = memory.device();
  QString spacing = '\n' + repeat(" ", 8);
  QString s = spacing + "#include <" + GPUtils::toDeviceName(data.name()) + ".h>";
  s +=  spacing + "processor " + GPUtils::toDeviceName(data.name()) + '\n';
  // memoryory config
  bool isDefault = true;
  for (uint k=0; k<data.nbWords(Pic::MemoryConfig); k++) {
    uint op = memory.normalizedWord(Pic::MemoryConfig, k);
    uint mask = data.config()._words[k].usedMask();
    if ( (op & mask)!=mask ) isDefault = false; // this is not completely correct but otherwise empty config is written...
  }
  if ( !isDefault ) {
    bool ok;
    s += spacing + generateDeviceConfig(memory, ok).stripWhiteSpace() + '\n';
  }
  // user ids
  QString tmp;
  for (uint k=0; k<data.nbWords(Pic::MemoryUserId); k++) {
    uint op = memory.normalizedWord(Pic::MemoryUserId, k);
    uint mask = data.range(Pic::MemoryUserId).rmask;
    if ( (op & mask)==mask ) continue;
    if ( data.architecture()==Pic::Arch_18X ) {
      uint ad = data.range(Pic::MemoryUserId).start + data.range(Pic::MemoryUserId).hexFileOffset + k*data.addressIncrement(Pic::MemoryUserId);
      s += spacing + "__IDLOCS " + toHexLabel(ad, data.nbCharsAddress()) + ", " + toHexLabel(op, data.nbCharsWord(Pic::MemoryUserId)) + '\n';
    } else tmp += toHex(op & 0xF, 1);
  }
  if ( !tmp.isEmpty() ) {
    s += spacing + "__IDLOCS 0x" + tmp + '\n';
  }
  // memoryory code
  tmp = QString::null;
  bool first = true, newOrg = true;
  uint nb = data.nbWords(Pic::MemoryCode);
  for (uint k=0; k<nb; k++) {
    uint op = memory.normalizedWord(Pic::MemoryCode, k);
    uint mask = data.range(Pic::MemoryCode).rmask;
    if ( (op & mask)==mask ) newOrg = true;
    else {
      if (newOrg) {
        if ( !first ) tmp += '\n';
        first = false;
        uint org = data.range(Pic::MemoryCode).start + data.range(Pic::MemoryCode).hexFileOffset + k*data.addressIncrement(Pic::MemoryCode);
        tmp += spacing + "org " + toHexLabel(org, data.nbCharsAddress());
        newOrg = false;
      }
      char buffer[512];
      buffer[0] = 0;
      uint op2 = ((k+1)<nb ? memory.word(Pic::MemoryCode, k+1) : 0);
      uint n = Coff::disassemble(op, op2, k, data.architecture(), buffer, 512);
      tmp += spacing + QString(buffer);
      if ( n==2 ) k++;
    }
  }
  if ( !tmp.isEmpty() ) s += "\n; memoryory code" + tmp + '\n';
  // eeprom data
  tmp = QString::null;
  first = true;
  newOrg = true;
  nb = data.nbWords(Pic::MemoryEeprom);
  for (uint k=0; k<nb; k++) {
    uint op = memory.normalizedWord(Pic::MemoryEeprom, k);
    uint mask = data.range(Pic::MemoryEeprom).rmask;
    if ( (op & mask)==mask ) newOrg = true;
    else {
      if (newOrg) {
        if ( !first ) tmp += '\n';
        first = false;
        uint org = data.range(Pic::MemoryEeprom).start + data.range(Pic::MemoryEeprom).hexFileOffset + k*data.addressIncrement(Pic::MemoryEeprom);
        tmp += spacing + "org " + toHexLabel(org, data.nbCharsAddress());
        newOrg = false;
      }
      tmp += spacing + "de " + toHexLabel(op, data.nbCharsWord(Pic::MemoryEeprom));
    }
  }
  if ( !tmp.isEmpty() ) s += "\n; memoryory eeprom" + tmp + '\n';
  s += spacing + "end\n";
  return s;
}
