/* Copyright (C) 2000 David Faure <faure@kde.org>, Alexander Neundorf <neundorf@kde.org>
                 2000, 2002 Carsten Pfeiffer <pfeiffer@kde.org>
   Copyright (C) 2006 Nicolas Hadacek <hadacek@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.
*/
#include "editlistbox.h"

#include <qstringlist.h>
#include <qlabel.h>

#include <kdialog.h>
#include <klocale.h>
#include <kapplication.h>
#include <knotifyclient.h>
#include <kiconloader.h>

//----------------------------------------------------------------------------
ListBox::ListBox(QWidget *parent, const char *name)
  : QListBox(parent, name)
{
  connect(this, SIGNAL(clicked(QListBoxItem *)), SLOT(clickedSlot(QListBoxItem *)));
}

void ListBox::clickedSlot(QListBoxItem *item)
{
  if ( item==0 ) clearSelection();
}

//----------------------------------------------------------------------------
EditListBox::EditListBox(QWidget *parent, const char *name, Mode mode, Buttons buttons)
  : QFrame(parent, name), _mode(mode), _buttons(buttons)
{
  init(0);
}

EditListBox::EditListBox(const CustomEditor& custom, QWidget *parent, const char *name, Mode mode, Buttons buttons)
  : QFrame(parent, name), _mode(mode), _buttons(buttons)
{
  m_lineEdit = custom.lineEdit();
  init(custom.representationWidget());
}

void EditListBox::init(QWidget *representationWidget)
{
    _addButton = 0;
    _removeButton = 0;
    _moveUpButton = 0;
    _moveDownButton = 0;
    _removeAllButton = 0;
    _resetButton = 0;
    setSizePolicy(QSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::MinimumExpanding));

    QGridLayout *grid = new QGridLayout(this, 1, 1, 0, KDialog::spacingHint());
    QHBoxLayout *hbox = new QHBoxLayout(KDialog::spacingHint());
    grid->addMultiCellLayout(hbox, 0,0, 0,1);
    KIconLoader loader;
    QIconSet iconset = loader.loadIcon("locationbar_erase", KIcon::Toolbar);
    KPushButton *button = new  KPushButton(iconset, QString::null, this);
    connect(button, SIGNAL(clicked()), SLOT(clearEdit()));
    hbox->addWidget(button);
    if ( representationWidget ) representationWidget->reparent( this, QPoint(0,0) );
    else m_lineEdit = new KLineEdit(this);
    hbox->addWidget(representationWidget ? representationWidget : m_lineEdit);
    m_listBox = new ListBox(this);
    grid->addWidget(m_listBox, 1, 0);
    _buttonsLayout = new QVBoxLayout(10);
    grid->addLayout(_buttonsLayout, 1, 1);

    setButtons(_buttons);

    connect(m_lineEdit,SIGNAL(textChanged(const QString&)),this,SLOT(typedSomething(const QString&)));
    m_lineEdit->setTrapReturnKey(true);
    connect(m_lineEdit,SIGNAL(returnPressed()),this,SLOT(addItem()));
    connect(m_listBox, SIGNAL(highlighted(int)), SLOT(enableMoveButtons(int)));

    // maybe supplied lineedit has some text already
    typedSomething( m_lineEdit->text() );
}

void EditListBox::setButtons(Buttons buttons)
{
    _buttons = buttons;

    delete _addButton;
    _addButton = 0;
    if ( buttons & Add ) {
        _addButton = new KPushButton(KStdGuiItem::add(), this);
        _addButton->setEnabled(false);
        _addButton->show();
        connect(_addButton, SIGNAL(clicked()), SLOT(addItem()));
        _buttonsLayout->addWidget(_addButton);
    }

    delete _removeButton;
    _removeButton = 0;
    if ( buttons & Remove ) {
        _removeButton = new KPushButton(KGuiItem(i18n("Remove"), "clear"), this);
        _removeButton->setEnabled(false);
        _removeButton->show();
        connect(_removeButton, SIGNAL(clicked()), SLOT(removeItem()));
        _buttonsLayout->addWidget(_removeButton);
    }

    delete _removeAllButton;
    _removeAllButton = 0;
    if ( buttons & RemoveAll ) {
        _removeAllButton = new KPushButton(KGuiItem(i18n("Remove All"), "delete"), this);
        _removeAllButton->show();
        connect(_removeAllButton, SIGNAL(clicked()), SLOT(clear()));
        _buttonsLayout->addWidget(_removeAllButton);
    }

    delete _moveUpButton;
    _moveUpButton = 0;
    delete _moveDownButton;
    _moveDownButton = 0;
    if ( buttons & UpDown ) {
        _moveUpButton = new KPushButton(KGuiItem(i18n("Move &Up"), "up"), this);
        _moveUpButton->setEnabled(false);
        _moveUpButton->show();
        connect(_moveUpButton, SIGNAL(clicked()), SLOT(moveItemUp()));
        _buttonsLayout->addWidget(_moveUpButton);
        _moveDownButton = new KPushButton(KGuiItem(i18n("Move &Down"), "down"), this);
        _moveDownButton->setEnabled(false);
        _moveDownButton->show();
        connect(_moveDownButton, SIGNAL(clicked()), SLOT(moveItemDown()));
        _buttonsLayout->addWidget(_moveDownButton);
    }

    delete _resetButton;
    _resetButton = 0;
    if ( buttons & Reset ) {
        _resetButton = new KPushButton(KStdGuiItem::reset(), this);
        _resetButton->show();
        connect(_resetButton, SIGNAL(clicked()), SIGNAL(reset()));
        _buttonsLayout->addWidget(_resetButton);
    }
}

void EditListBox::typedSomething(const QString& text)
{
    QListBoxItem *item = m_listBox->selectedItem();
    if (item) {
        if( m_listBox->currentText()!= m_lineEdit->text() )
        {
            // IMHO changeItem() shouldn't do anything with the value
            // of currentItem() ... like changing it or emitting signals ...
            // but TT disagree with me on this one (it's been that way since ages ... grrr)
            bool block = m_listBox->signalsBlocked();
            m_listBox->blockSignals( true );
            m_listBox->changeItem(text, m_listBox->index(item));
            m_listBox->blockSignals( block );
            emit changed();
        }
    }
    updateButtons();
}

void EditListBox::moveItemUp()
{
    if (!m_listBox->isEnabled())
    {
        KNotifyClient::beep();
        return;
    }

    const unsigned int selIndex = m_listBox->currentItem();
    if (selIndex == 0)
    {
        KNotifyClient::beep();
        return;
    }

    QListBoxItem *selItem = m_listBox->item(selIndex);
    m_listBox->takeItem(selItem);
    m_listBox->insertItem(selItem, selIndex-1);
    m_listBox->setCurrentItem(selIndex - 1);

    updateButtons();
    emit changed();
}

void EditListBox::moveItemDown()
{
    if (!m_listBox->isEnabled())
    {
        KNotifyClient::beep();
        return;
    }

    unsigned int selIndex = m_listBox->currentItem();
    if (selIndex == m_listBox->count() - 1)
    {
        KNotifyClient::beep();
        return;
    }

    QListBoxItem *selItem = m_listBox->item(selIndex);
    m_listBox->takeItem(selItem);
    m_listBox->insertItem(selItem, selIndex+1);
    m_listBox->setCurrentItem(selIndex + 1);

    updateButtons();
    emit changed();
}

void EditListBox::addItem()
{
    // when m_checkAtEntering is true, the add-button is disabled, but this
    // slot can still be called through Key_Return/Key_Enter. So we guard
    // against this.
    if ( !_addButton || !_addButton->isEnabled() ) return;

    const QString& currentTextLE=m_lineEdit->text();
    bool alreadyInList(false);
    //if we didn't check for dupes at the inserting we have to do it now
    if ( _mode==DuplicatesDisallowed )
    {
        // first check current item instead of dumb iterating the entire list
        if ( m_listBox->currentText() == currentTextLE )
            alreadyInList = true;
        else
        {
            StringComparisonMode mode = (StringComparisonMode) (ExactMatch | CaseSensitive );
            alreadyInList =(m_listBox->findItem(currentTextLE, mode) );
        }
    }

    bool block = m_lineEdit->signalsBlocked();
    m_lineEdit->blockSignals(true);
    m_lineEdit->clear();
    m_lineEdit->blockSignals(block);

    m_listBox->setSelected(m_listBox->selectedItem(), false);

    if (!alreadyInList)
    {
        block = m_listBox->signalsBlocked();
        m_listBox->blockSignals( true );
        m_listBox->insertItem(currentTextLE);
        m_listBox->blockSignals( block );
        emit changed();
	emit added( currentTextLE );
    }
    updateButtons();
}

void EditListBox::clearEdit()
{
  m_listBox->clearSelection();
  m_lineEdit->clear();
  m_lineEdit->setFocus();
  updateButtons();
}

void EditListBox::removeItem()
{
    int selected = m_listBox->currentItem();

    if ( selected >= 0 )
    {
	QString removedText = m_listBox->currentText();

        m_listBox->removeItem( selected );
        if ( count() > 0 )
            m_listBox->setSelected( QMIN( selected, count() - 1 ), true );

        emit changed();
	emit removed( removedText );
    }
    updateButtons();
}

void EditListBox::enableMoveButtons(int)
{
    m_lineEdit->setText(m_listBox->currentText());
    updateButtons();
}

void EditListBox::clear()
{
    m_listBox->clear();
    m_lineEdit->clear();
    m_lineEdit->setFocus();
    updateButtons();
    emit changed();
}

void EditListBox::insertStringList(const QStringList& list, int index)
{
    m_listBox->insertStringList(list,index);
    updateButtons();
}

QStringList EditListBox::items() const
{
    QStringList list;
    for (QListBoxItem const * i = m_listBox->firstItem(); i != 0; i = i->next() )
	list.append( i->text());
    return list;
}

void EditListBox::setItems(const QStringList& items)
{
  m_listBox->clear();
  m_listBox->insertStringList(items, 0);
  m_lineEdit->clear();
  updateButtons();
}

void EditListBox::updateButtons()
{
  int index = m_listBox->index(m_listBox->selectedItem());
  QString text = m_lineEdit->text();
  if (_addButton) {
    if ( _mode!=DuplicatesCheckedAtEntering ) _addButton->setEnabled(!text.isEmpty());
    else if ( text.isEmpty() ) _addButton->setEnabled(false);
    else {
      StringComparisonMode mode = (StringComparisonMode) (ExactMatch | CaseSensitive );
      _addButton->setEnabled(!m_listBox->findItem(text, mode));
    }
  }
  if (_removeButton) _removeButton->setEnabled(index!=-1);
  if (_moveUpButton) _moveUpButton->setEnabled(index>0);
  if (_moveDownButton) _moveDownButton->setEnabled(index!=-1 && index<(int(m_listBox->count())-1));
}

//----------------------------------------------------------------------------
EditListBox::CustomEditor::CustomEditor( KComboBox *combo )
{
    m_representationWidget = combo;
    m_lineEdit = dynamic_cast<KLineEdit*>( combo->lineEdit() );
    Q_ASSERT( m_lineEdit );
}
