/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *             (C) 2003 by Alain Gibaud                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "hex_buffer.h"

#include <qtextstream.h>
#include <qfile.h>

#include "common/global/global.h"
#include "devices/base/generic_device.h"

//-----------------------------------------------------------------------------
void HexBuffer::set(uint k, uint value)
{
  std::map<uint, uint>::operator[](k) = value;
}

uint *HexBuffer::operator+(uint k)
{
  iterator i = find(k);
  if ( i!=end() ) return &(*i).second;
  set(k, Device::INVALID);
  return &std::map<uint, uint>::operator[](k);
}

uint HexBuffer::operator[](uint k) const
{
  const_iterator i = find(k);
  if ( i!=end() ) return (*i).second;
  return Device::INVALID;
}

//-----------------------------------------------------------------------------
const char * const HexBuffer::FORMATS[Nb_Formats] = {
  "inhx8m", /*"inhx8s", */"inhx16", "inhx32"
};

void HexBuffer::savePartial(QTextStream &stream, Format format) const
{
  uint oldseg = Device::INVALID;
  const_iterator block =  begin();
  int len;
  while ( fetchNextBlock(block, end(), &len) ) {
    // block found, write it
    uint seg = (*block).first >> 15 ; // 2 * seg address
    if ( format == IHX32 && seg != oldseg ) {
      char buf[50];
      unsigned int check = 0x02 + 0x04 + (seg >> 8) + (seg & 0xFF);
      sprintf(buf,":02000004%04X%02X\n", seg, -check & 0xFF );
      stream << buf;
      oldseg = seg ;
    }
    writeHexBlock(stream, len, block, format);
  }
}

void HexBuffer::saveEnd(QTextStream &stream)
{
  stream << ":00000001FF\n";
}

/* Write one line of Intel-hex file
 * Original code source from Timo Rossi,
 * modified by Alain Gibaud to support large blocks write
 */
void HexBuffer::writeHexBlock(QTextStream &stream,
                          int reclen, /* length (in words) */
                          const_iterator& data, /* pointer to 1st data word (incremented by function)*/
                          Format format)
{
    while ( reclen>HEXBLKSIZE ) {
      writeHexBlock(stream, HEXBLKSIZE, data, format);
      reclen -= HEXBLKSIZE;
    }
    if ( reclen<=0 ) return; /* Oops, block has just a HEXBLKSIZE * n size */

    char buf[20];
    int check = 0;

    // line start
    uint loc =  (*data).first;
    switch (format) {
      case IHX8M:
      case IHX32:
        loc *= 2;
        sprintf(buf,  ":%02X%04X00", 2*reclen, loc & 0xFFFF);
        check += ((loc) & 0xff) + (((loc) >> 8) & 0xff) + 2*reclen;
        break;
      case IHX16:
        sprintf(buf, ":%02X%04X00", reclen, loc & 0xFFFF);
        check += (loc & 0xff) + ((loc >> 8) & 0xff) + reclen;
      case Nb_Formats: Q_ASSERT(false); break;
    }
    stream << buf;

    // data
    for (; reclen > 0; ++data, --reclen) {
      uint word =  (*data).second;
      switch (format) {
        case IHX8M:
        case IHX32:
          sprintf( buf, "%02X%02X", (int)(word & 0xff), (int)((word >> 8) & 0xff));
          break;
        case IHX16:
          sprintf(buf, "%02X%02X", (int)((word >> 8) & 0xff), (int)(word & 0xff));
          break;
        case Nb_Formats: Q_ASSERT(false); break;
      }
      stream << buf;
      check += (word & 0xff) + ((word >> 8) & 0xff);
    }

    // checksum, assumes 2-complement
    sprintf(buf, "%02X\n", (-check) & 0xff);
    stream << buf;
}

/* -------------------------------------------------------------------------
    This routine detects the next block to output
    A block is a set of consecutive addresse words not
    containing 0xFFFFFFFF
    @return true if a block has been detected
    'it' is updated to point to the first address of the block
    '*len' contains the size of the block
*/
bool HexBuffer::fetchNextBlock(const_iterator& it, const const_iterator &end, int *len)
{
    uint startadr, curadr;
    // for( i = *start ; (i < MAXPICSIZE) && (Mem[i] == INVALID) ; ++i) ;
    // skip non-used words

    // Search block start
    while ( it!=end && (*it).second==Device::INVALID ) ++it;

    // if(i >= MAXPICSIZE ) return false ;
    if ( it==end ) return false;

    //for( *start = i ; (i < MAXPICSIZE) && (Mem[i] != INVALID) ; ++i) ;
    //*len = i - *start ;
    // search block end - a block may not cross a segment boundary
    const_iterator itt(it) ;
    for (curadr = startadr = (*itt).first, ++itt; itt!=end; ++itt) {
        if ( (*itt).first!=curadr+1 ) break;    // non contiguous addresses
        if ( (*itt).second==Device::INVALID ) break;   // unused word found
        if ( (((*itt).first) & 0xFFFF0000U)!=(curadr & 0xFFFF0000U) ) break; // cross segment boundary

        curadr =   (*itt).first  ;
    }
    *len = curadr - startadr + 1 ;

    return *len != 0  ;
}

bool HexBuffer::load(QFile &file, QString &message)
{
  clear();
  switch (readHex(file)) {
    case NoError: break;
    case CRCError: message = i18n("CRC mismatch in hex file."); return false;
    case UnrecognizedError: message = i18n("Unrecognized hex file."); return false;
  }
  return true;
}

/* -------------------------------------------------------------------------
   Read a Intel HEX file of either INHX16 or INHX8M format type, detecting
   the format automagicly by the wordcount and length of the line
   Tested in 8 and 16 bits modes
   ------------------------------------------------------------------------ */
HexBuffer::ReadError HexBuffer::readHex(QFile &file)
{
    /* Read each line of the file */
    uint line = 0, addrH = 0; // upper 16 bits of 32 bits address  (inhx32 format)
    bool is_inhx32 = false;
    char buf[1024];
    while ( file.readLine(buf, 1023)!=-1 ) {
        line++;

        const char *p = buf;
        /* If it isn't a valid intel hex line, then move to the next line */
        if( *p != ':' ) continue;

        /* because we need the exact line length for format detection,
           strip out CR/LF (and any non-hex digit !) */
        uint kend;
        for (kend = strlen(buf)-1; !isxdigit(buf[kend]); --kend) ;
        if ( buf[kend]==':' ) continue; // ':' is alone
        buf[kend+1] = 0;
        uint bytecount = (kend-9) / 2; // number of data bytes of this record

        /* Get the byte count, the address and the type for this line.
           Because the checksum is computed from these elements,
           I assume they are mandatory, thus I can read them together */
        uint count, addr, type;
        if( sscanf( ++p, "%02X%04X%02X", &count , &addr, &type) != 3) return UnrecognizedError;
        /* Point to the first data byte */
        p += 8;

        /* Start calculating the checksum */
        uint cksum = count + (addr >> 8) + (addr & 0xFF) + type;

        /* End of file field ?, exit now */
        if( type==0x01 ) {
            uint data;
            if ( sscanf(p, "%02X", &data)!=1 ) return UnrecognizedError; // bad EOF record
            if ( ((cksum+data) & 0xFF)!=0 ) return CRCError; // bad CRC
            return NoError;  // OK, EOF
        }
        // example : linear extended record for addr 0x21xxxx
        // (blanks inserted for clarify)
        // :02 0000 04 0021 D9
        if ( type==0x04 ) {
            uint data;
            if( sscanf(p, "%04X%02X", &addrH , &data)!=2 ) return UnrecognizedError; // bad address record
            cksum += (addrH & 0xFF);
            cksum += (addrH >> 8);
            is_inhx32 = true;
            if( (cksum + data) & 0xFF!=0 ) return CRCError; // bad CRC
            //qDebug("new address high: %s", toHex(addrH<<16, 8).data());
            continue; // goto next record
        }

        /* Figure out if its INHX16 or INHX8M
            if count is a 16 bits words count => INHX16
            if count is a byte count => INHX8M or INHX32 */
        if ( bytecount==count ) {
            /* Processing a INHX8M line */
            /* Modified to be able to read fuses from hexfile created by C18 toolchain */
            /* changed by Tobias Schoene 9 April 2005, */
            /* modified by A.G, because low and hi bytes was swapped in Tobias's code , 8 may 2005
            */
            uint addrbase = ((addrH << 16) | addr);
            //qDebug("line %i: address %s", line, toHex(addrbase, 8).data());
            for (uint x = 0; x<count; x++) {
              uint data;
              if ( sscanf(p, "%02X", &data)!=1 ) return UnrecognizedError;
              p += 2;
              // A.G: I suspect possible initialization problem
              // if block begins at odd address
              // because |= works on an uninitalized word
              // however, I  don't know if such a situation can occurs
              uint a = addrbase+x >> 1;
              uint value = (*this)[a];
              if ( addrbase+x & 1 ) {
                value &= 0xFFU;            // clear out hi bits (patch from Chris Teague chris.teague@gmail.com)
                set(a, value | data << 8); // Odd addr => Hi byte
              } else {
                value &= ~0xFFU;      // clear out low bits (patch from Chris Teague chris.teague@gmail.com)
                set(a, value | data); // Low byte
              }
              //if ( x==0 ) qDebug("fb@%s: %s", toHex(addrbase+x >> 1, 8).data(), toHex(fb[addrbase+x >> 1], 8).data());
              cksum += data;
            }
        } else if ( bytecount==count*2 ) {
            if (is_inhx32) return UnrecognizedError; // inconsistant format
            /* Processing a INHX16 line */
            for(uint x = 0; x < count; x++ ) {
                uint datal, datah;
                if( sscanf(p, "%02X%02X", &datah, &datal)!=2 ) return UnrecognizedError;
                p += 4;
                set(addr+x, datal | (datah << 8));
                cksum += datah;
                cksum += datal;
            }
        } else return UnrecognizedError; // Brrrr !! Strange format.

        /* Process the checksum */
        uint data;
        if( sscanf(p, "%02X", &data)!=1 ) return UnrecognizedError; // fmt error
        if( ((data + cksum) & 0xFF)!=0 ) return CRCError;
    }

    return UnrecognizedError;
}
