/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "pic.h"

#include "common/global/global.h"
#include "common/common/misc.h"
#include "pic_register.h"
#include "pic_config.h"

//------------------------------------------------------------------------------
const Pic::MemoryRangeTypeData Pic::MEMORY_RANGE_TYPE_DATA[Nb_MemoryRangeTypes] = {
  { "program",            I18N_NOOP("Code memory"),         Writable },
  { "calibration",        I18N_NOOP("Calibration"),         Writable },
  { "user_ids",           I18N_NOOP("User IDs"),            Writable },
  { "device_id",          I18N_NOOP("Device ID"),           ReadOnly },
  { "config",             I18N_NOOP("Configuration Bits"),  Writable },
  { "eeprom",             I18N_NOOP("Data EEPROM"),         Writable },
  { "debug_vector",       I18N_NOOP("Debug Vector"),        Writable },
  { "hardware_stack",     I18N_NOOP("Hardware Stack"),      ReadOnly },
  { "calibration_backup", I18N_NOOP("Calibration Backup"),  Writable },
  { "program_executive",  I18N_NOOP("Program Executive"),   Writable }
};

const Pic::ArchitectureData Pic::ARCHITECTURE_DATA[Nb_Architectures] = {
//  name PC nbBytes packed codeAlignment nbBitsRegister registerBankLength
//  {Code, Cal, UserID, DevId, Conf, EEPROM, DebugVec, HardStack, CalBackup, Program Executive}
  { "10X", I18N_NOOP("Baseline Family"),  9, 2, false, 16,  8, 0x020, { 12, 12, 12, 12, 12,  8, 12,  0, 12,  0 } }, // 9, 10, 11 or 12-bit program counter
  { "16X", I18N_NOOP("Midrange Family"), 13, 2, false, 16,  8, 0x080, { 14, 14, 14, 14, 14,  8, 14,  0, 14,  0 } }, // max eeprom: 256 words
  { "17X", I18N_NOOP("17X Family"),      21, 2, false, 16,  8, 0x100, { 16,  0,  0,  0, 16,  8,  0,  0,  0,  0 } },
  { "18X", I18N_NOOP("18X Family"),      21, 2, true,  16,  8, 0x100, { 16,  8,  8,  8,  8,  8, 16,  0,  8,  0 } },
  { "30X", I18N_NOOP("dsPICs Family"),   23, 4, false, 16, 16, 0xA00, { 24,  0,  0, 16, 16, 16, 24,  0,  0, 24 } } // user id == unitid 24bits?   // dsPIC: eeprom max = 2 kwords = 4 kbytes
};

const Pic::ProgVoltageTypeData Pic::PROG_VOLTAGE_TYPE_DATA[Nb_ProgVoltageTypes] = {
  { "vpp"            },
  { "vdd_prog"       },
  { "vdd_prog_write" },
  { "vdd_prog_read"  }
};

//-----------------------------------------------------------------------------
Pic::Data::Data()
{
  for (uint i=0; i<Nb_ProgVoltageTypes; i++) {
    _voltages[i].min = 0.0;
    _voltages[i].max = 0.0;
    _voltages[i].nominal = 0.0;
  }
  for (uint i=0; i<Nb_MemoryRangeTypes; i++) {
    _ranges[i].properties = NotPresent;
    _ranges[i].start = 0;
    _ranges[i].end = 0;
    _ranges[i].mask = 0;
    _ranges[i].rmask = 0;
    _ranges[i].hexFileOffset = 0;
  }
  _config = new Config(*this);
  _calibration.opcode = 0;
  _calibration.opcodeMask = 0;
  _registersData = new RegistersData(*this);
}

Pic::Data::~Data()
{
  delete _config;
}

QString Pic::Data::informationUrl() const
{
  QString url = "http://microchip.com/stellent/idcplg?IdcService=SS_GET_PAGE&nodeId=";
  if ( document().isEmpty() ) {
    url += "2044&AllWords=";
    if ( name().startsWith("PS")) url += name();
    if ( name().startsWith("30") ) url += "dsPIC" + name();
    else url += "PIC" + name();
  } else url += "1335&dDocName=en" + document();
  return url;
}

QString Pic::Data::informationHtml() const
{
  // memory type
  QString s = htmlTableRow(i18n("Memory Type"), i18n(Device::MEMORY_TECHNOLOGY_DATA[memoryTechnology()].label));
  if ( isPresent(Pic::MemoryCode) ) {
    uint nbw = nbWords(Pic::MemoryCode);
    QString tmp = i18n("%1 words").arg(formatNumber(nbw));
    tmp += i18n(" (%2 bits)").arg(nbBitsWord(Pic::MemoryCode));
    s += htmlTableRow(i18n(Pic::MEMORY_RANGE_TYPE_DATA[Pic::MemoryCode].label), tmp);
  }
  if ( isPresent(Pic::MemoryEeprom) ) {
    uint nbw = nbWords(Pic::MemoryEeprom);
    QString tmp = i18n("%1 bytes").arg(formatNumber(nbw));
    tmp += i18n(" (%2 bits)").arg(nbBitsWord(Pic::MemoryEeprom));
    if ( !isWritable(Pic::MemoryEeprom) ) tmp += i18n(" (not programmable)");
    s += htmlTableRow(i18n(Pic::MEMORY_RANGE_TYPE_DATA[Pic::MemoryEeprom].label), tmp);
  }
  // io ports
  QString tmp;
  uint nb = 0;
  for (uint i=0; i<Device::MAX_NB_PORTS; i++) {
    if ( !registersData().hasPort(i) ) continue;
    uint nbBits = 0;
    for (uint k=0; k<Device::MAX_NB_PORT_BITS; k++) if ( registersData().hasPortBit(i, k) ) nbBits++;
    tmp += registersData().portName(i) + "[" + QString::number(nbBits) + "] ";
    nb++;
  }
  if ( nb==0 ) tmp = i18n("<none>");
  s += htmlTableRow(i18n("IO Ports"), tmp);
  return s;
}

bool Pic::Data::isWritable(MemoryRangeType type) const
{
  return ( MEMORY_RANGE_TYPE_DATA[type].properties & Pic::Writable
           && (range(type).properties & Programmable) );
}

uint Pic::Data::addressIncrement(MemoryRangeType type) const
{
  uint inc = ARCHITECTURE_DATA[_architecture].nbBytesWord;
  if ( ARCHITECTURE_DATA[_architecture].packed && type==MemoryCode ) return inc;
  return inc / 2;
}

uint Pic::Data::nbWords(MemoryRangeType type) const
{
  if ( !isPresent(type) ) return 0;
  return nbAddresses(type) / addressIncrement(type);
}

uint Pic::Data::nbAddresses(MemoryRangeType type) const
{
  if ( !isPresent(type) ) return 0;
  return (range(type).end - range(type).start + 1);
}

QString Pic::Data::fname(Device::FrequencyRange::Special special) const
{
  QString s = name();
  switch (special) {
    case Device::FrequencyRange::Normal: break;
    case Device::FrequencyRange::LowPower:
      // assume name is of form "NNX..."
      s.insert(2, 'L');
      break;
    case Device::FrequencyRange::LowVoltage:
      // assume name is of form "NNXN..."
      s.replace(2, 1, "LV");
      break;
    case Device::FrequencyRange::Nb_Specials: Q_ASSERT(false); break;
  }
  return s;
}

bool Pic::Data::matchId(uint rawId, uint deviceIdAddress, Device::IdData &idata) const
{
  if ( !isPresent(MemoryDeviceId) || deviceIdAddress!=range(MemoryDeviceId).start )
    return false;
  uint nid = 0;
  switch ( architecture() ) {
    case Arch_10X:
    case Arch_16X:
    case Arch_17X:
    case Arch_18X:
      nid = rawId & ~0x1F;
      idata.revision = rawId & 0x1F;
      break;
    case Arch_30X:
      nid = (rawId >> 16) & 0xFFFF;
      idata.revision = (rawId >> 6) & 0x3F;
      idata.minorRevision = rawId & 0x3F;
      idata.process = (rawId >> 12) & 0xF;
      break;
    case Nb_Architectures: Q_ASSERT(false); break;
  }
  return ( nid==id() );
}

QStringList Pic::Data::idNames(const Device::IdList &idList) const
{
  QStringList list;
  QMap<QString, Device::IdData>::const_iterator it = idList.begin();
  for (; it!=idList.end(); ++it) {
    switch (_architecture) {
      case Arch_10X:
      case Arch_16X:
      case Arch_17X:
      case Arch_18X:
        list += i18n("%1 (rev. %2)").arg(it.key()).arg(it.data().revision);
        break;
      case Arch_30X:
        list += i18n("%1 (proc. %2; rev. %3.%4)").arg(it.key())
                .arg(it.data().process).arg(it.data().revision).arg(it.data().minorRevision);
        break;
      case Nb_Architectures: Q_ASSERT(false); break;
    }
  }
  return list;
}

bool Pic::Data::checkCalibration(const Device::Array &data, QString *message) const
{
  Q_ASSERT( nbWords(MemoryCal)==data.count() );
  for (uint i=0; i<data.count(); i++) {
    QString address = toHexLabel(range(MemoryCal).start + i*addressIncrement(MemoryCal), nbCharsAddress());
    if ( data[i]==range(MemoryCal).mask ) {
      if (message) *message = i18n("Calibration word at address %1 is blank.").arg(address);
      return false;
    }
  }
  if ( data.count()==1 ) {
    if ( (data[0] & _calibration.opcodeMask)!=_calibration.opcode ) {
      if (message) *message = i18n("Calibration word is not a compatible opcode (%2).")
                              .arg(toHexLabel(_calibration.opcode, nbCharsWord(MemoryCode)));
      return false;
    }
  }
  return true;
}

uint Pic::Data::nbWordsAlignment(MemoryRangeType type) const
{
  if ( type!=MemoryCode ) return 1;
  return ARCHITECTURE_DATA[architecture()].nbWordsCodeAlignment;
}

const Pic::RegistersData &Pic::Data::registersData() const
{
  return static_cast<const RegistersData &>(*_registersData);
}

//----------------------------------------------------------------------------
QDataStream &Pic::operator <<(QDataStream &s, const VoltageData &vd)
{
  s << vd.min << vd.max << vd.nominal;
  return s;
}
QDataStream &Pic::operator >>(QDataStream &s, VoltageData &vd)
{
  s >> vd.min >> vd.max >> vd.nominal;
  return s;
}

QDataStream &Pic::operator <<(QDataStream &s, const MemoryRangeData &mrd)
{
  s << Q_UINT8(mrd.properties) << mrd.start << mrd.end << mrd.mask << mrd.rmask << mrd.hexFileOffset;
  return s;
}
QDataStream &Pic::operator >>(QDataStream &s, MemoryRangeData &mrd)
{
  Q_UINT8 properties;
  s >> properties >> mrd.start >> mrd.end >> mrd.mask >> mrd.rmask >> mrd.hexFileOffset;
  mrd.properties = MemoryRangeProperties(properties);
  return s;
}

QDataStream &Pic::operator <<(QDataStream &s, const ChecksumData &cd)
{
  s << cd.constant << cd.protectedConstant << Q_UINT8(cd.protectedType);
  return s;
}
QDataStream &Pic::operator >>(QDataStream &s, ChecksumData &cd)
{
  Q_UINT8 type;
  s >> cd.constant >> cd.protectedConstant >> type;
  cd.protectedType = ChecksumType(type);
  return s;
}

QDataStream &Pic::operator <<(QDataStream &s, const CalibrationData &cd)
{
  s << cd.opcode << cd.opcodeMask;
  return s;
}
QDataStream &Pic::operator >>(QDataStream &s, CalibrationData &cd)
{
  s >> cd.opcode >> cd.opcodeMask;
  return s;
}

QDataStream &Pic::operator <<(QDataStream &s, const Data &data)
{
  s << static_cast<const Device::Data &>(data);
  s << data._id;
  for (uint i=0; i<Nb_ProgVoltageTypes; i++) s << data._voltages[i];
  for (uint i=0; i<Nb_MemoryRangeTypes; i++) s << data._ranges[i];
  s << *data._config << data._checksum << data._calibration << static_cast<const Pic::RegistersData &>(*data._registersData);
  return s;
}
QDataStream &Pic::operator >>(QDataStream &s, Data &data)
{
  s >> static_cast<Device::Data &>(data);
  s >> data._id;
  for (uint i=0; i<Nb_ProgVoltageTypes; i++) s >> data._voltages[i];
  for (uint i=0; i<Nb_MemoryRangeTypes; i++) s >> data._ranges[i];
  s >> *data._config >> data._checksum >> data._calibration >> static_cast<Pic::RegistersData &>(*data._registersData);
  return s;
}
