/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#ifndef PIC_H
#define PIC_H

#include "common/global/global.h"
#include <qstringlist.h>
#include <qmap.h>

#include "devices/base/generic_device.h"

namespace Pic
{
class XmlToData;
class Group;
class Config;
class RegistersData;

//-----------------------------------------------------------------------------
struct VoltageData {
  double min, max, nominal;
};
enum ProgVoltageType { Vpp = 0, VddBulkErase, VddWrite, VddRead, Nb_ProgVoltageTypes };
struct ProgVoltageTypeData {
  const char *name;
};
extern const ProgVoltageTypeData PROG_VOLTAGE_TYPE_DATA[Nb_ProgVoltageTypes];

struct CalibrationData {
  uint opcode, opcodeMask;
};

enum MemoryRangeType { MemoryCode = 0, MemoryCal, MemoryUserId, MemoryDeviceId,
                       MemoryConfig, MemoryEeprom, MemoryDebugVector, MemoryHardwareStack,
                       MemoryCalBackup, MemoryProgramExecutive, Nb_MemoryRangeTypes };
enum MemoryRangeTypeProperty { ReadOnly = 0, Writable = 1 };
Q_DECLARE_FLAGS(MemoryRangeTypeProperties, MemoryRangeTypeProperty)
Q_DECLARE_OPERATORS_FOR_FLAGS(MemoryRangeTypeProperties)
struct MemoryRangeTypeData {
  const char *name, *label;
  MemoryRangeTypeProperties properties;
};
extern const MemoryRangeTypeData MEMORY_RANGE_TYPE_DATA[Nb_MemoryRangeTypes];

enum Architecture { Arch_10X = 0, Arch_16X, Arch_17X, Arch_18X, Arch_30X, Nb_Architectures };
struct ArchitectureData {
  const char *name, *label;
  uint nbBitsPC;                     // nb bits max for PC
  uint nbBytesWord;                  // nb bytes per word (hex file and icd2)
  bool packed;
  uint nbWordsCodeAlignment;         // alignment for programming code memory
  uint nbBitsRegister;
  uint registerBankLength;
  uint nbBits[Nb_MemoryRangeTypes];  // nb bits per word
};
extern const ArchitectureData ARCHITECTURE_DATA[Nb_Architectures];

enum MemoryRangeProperty { NotPresent = 0, Present = 1, Programmable = 2 };
Q_DECLARE_FLAGS(MemoryRangeProperties, MemoryRangeProperty)
Q_DECLARE_OPERATORS_FOR_FLAGS(MemoryRangeProperties)
struct MemoryRangeData {
  MemoryRangeProperties properties;
  uint start, end, mask, rmask, hexFileOffset;
};

enum ChecksumType { NormalChecksum, XOR4Checksum, XNOR7Checksum, XNOR8Checksum };
struct ChecksumData
{
  uint constant, protectedConstant;
  ChecksumType protectedType;
};

//-----------------------------------------------------------------------------
class Data : public Device::Data
{
public:
  Data();
  virtual ~Data();
  virtual QString fname(Device::FrequencyRange::Special special) const;
  virtual QString informationUrl() const;
  virtual QString informationHtml() const;
  uint id() const { return _id; }
  bool isPresent(MemoryRangeType type) const { return (range(type).properties & Present); }
  bool isWritable(MemoryRangeType type) const;
  uint nbAddresses(MemoryRangeType type) const;
  uint nbWords(MemoryRangeType type) const;
  uint addressIncrement(MemoryRangeType type) const;
  uint nbWordsAlignment(MemoryRangeType type) const;
  const MemoryRangeData &range(MemoryRangeType type) const { return _ranges[type]; }
  virtual uint nbBitsAddress() const { return ARCHITECTURE_DATA[_architecture].nbBitsPC; }
  uint nbBitsWord(MemoryRangeType type) const { return ARCHITECTURE_DATA[_architecture].nbBits[type]; }
  uint nbBytesWord(MemoryRangeType type) const { return nbBitsToNbBytes(nbBitsWord(type)); }
  uint nbCharsWord(MemoryRangeType type) const { return nbBitsToNbChars(nbBitsWord(type)); }
  const Config &config() const { return *_config; }
  Architecture architecture() const { return Architecture(_architecture); }
  const VoltageData &voltage(ProgVoltageType type) const { return _voltages[type]; }
  virtual bool canWriteCalibration() const { return isPresent(MemoryCal) && isWritable(MemoryCal); }
  bool checkCalibration(const Device::Array &data, QString *message = 0) const;
  bool matchId(uint rawId, uint deviceIdAddress, Device::IdData &data) const;
  QStringList idNames(const Device::IdList &idList) const;
  const ChecksumData &checksumData() const { return _checksum; }
  const RegistersData &registersData() const;

private:
  uint _id;
  VoltageData      _voltages[Nb_ProgVoltageTypes];
  MemoryRangeData  _ranges[Nb_MemoryRangeTypes];
  Config          *_config;
  ChecksumData     _checksum;
  CalibrationData  _calibration;

  friend class XmlToData;
  friend class Group;
  friend QDataStream &operator <<(QDataStream &s, const Data &data);
  friend QDataStream &operator >>(QDataStream &s, Data &data);
};

QDataStream &operator <<(QDataStream &s, const VoltageData &vd);
QDataStream &operator >>(QDataStream &s, VoltageData &vd);
QDataStream &operator <<(QDataStream &s, const MemoryRangeData &mrd);
QDataStream &operator >>(QDataStream &s, MemoryRangeData &mrd);
QDataStream &operator <<(QDataStream &s, const ChecksumData &cd);
QDataStream &operator >>(QDataStream &s, ChecksumData &cd);
QDataStream &operator <<(QDataStream &s, const CalibrationData &cd);
QDataStream &operator >>(QDataStream &s, CalibrationData &cd);
QDataStream &operator <<(QDataStream &s, const Data &data);
QDataStream &operator >>(QDataStream &s, Data &data);

} // namespace

#endif
