/***************************************************************************
 *   Copyright (C) 2006 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "config_gen.h"

#include <qlayout.h>
#include <qlabel.h>
#include <qcombobox.h>
#include <klocale.h>

#include "device_gui.h"
#include "devices/base/generic_memory.h"
#include "devices/gui/hex_view.h"
#include "devices/gui/memory_editor.h"
#include "devices/base/device_group.h"
#include "devices/gui/device_group_ui.h"
#include "devices/list/device_list.h"
#include "text_editor.h"
#include "tools/base/tool_list.h"

ConfigGenerator::ConfigGenerator(const Device::Data *data, const Tool::Group &group, PURL::ToolType stype, QWidget *parent)
  : Dialog(parent, "config_generator", true, i18n("Config Generator"), Close|User1, Close, false, QSize(400, 300)),
    _memory(0), _configEditor(0)
{
  QVBoxLayout *top = new QVBoxLayout(mainWidget(), 10, 10);

  QHBoxLayout *hbox = new QHBoxLayout(top);
  QLabel *label = new QLabel(i18n("Device:"), mainWidget());
  hbox->addWidget(label);
  _deviceChooser = new DeviceChooser::Button(false, mainWidget());
  connect(_deviceChooser, SIGNAL(changed()), SLOT(set()));
  hbox->addWidget(_deviceChooser);
  hbox->addSpacing(20);

  label = new QLabel(i18n("Config Type:"), mainWidget());
  hbox->addWidget(label);
  _configType = new QComboBox(mainWidget());
  connect(_configType, SIGNAL(activated(int)), SLOT(typeChanged()));
  for (uint i=0; i<Tool::lister().nbGroups(); i++)
    _configType->insertItem(Tool::lister().group(i).label());
  hbox->addWidget(_configType);

  label = new QLabel(i18n("Tool Type:"), mainWidget());
  hbox->addWidget(label);
  _toolType = new QComboBox(mainWidget());
  for (uint i=0; i<PURL::Nb_ToolTypes; i++)
    _toolType->insertItem(i18n(PURL::TOOL_TYPE_LABELS[i]));
  _toolType->setMinimumWidth(_toolType->sizeHint().width());
  connect(_toolType, SIGNAL(activated(int)), SLOT(computeConfig()));
  hbox->addWidget(_toolType);
  hbox->addStretch(1);

  _hbox = new QHBoxLayout(top);

  _text = new SimpleTextEditor(Editor::NoCapability, PURL::Nb_FileTypes, mainWidget());
  _text->setReadOnly(true);
  top->addWidget(_text);

  setButtonText(User1, i18n("Copy to clipboard"));

  set(data, group, stype);
}

ConfigGenerator::~ConfigGenerator()
{
  delete _memory;
}

void ConfigGenerator::set(const Device::Data *data, const Tool::Group &group, PURL::ToolType stype)
{
  QString device;
  if ( data==0 ) {
    QValueVector<QString> devices = group.supportedDevices();
    if ( devices.isEmpty() ) return;
    device = devices[0];
  } else device = data->name();
  _deviceChooser->setDevice(device);
  _configType->setCurrentItem(Tool::lister().groupIndex(group.name()));
  setToolType(stype);
  set();
}

void ConfigGenerator::typeChanged()
{
  setToolType(PURL::Nb_ToolTypes);
  computeConfig();
}

void ConfigGenerator::setToolType(PURL::ToolType stype)
{
  uint group = _configType->currentItem();
  _toolType->clear();
  uint k = 0;
  for (uint i=0; i<PURL::Nb_ToolTypes; i++) {
    if ( Tool::lister().group(group).sourceFileType(PURL::ToolType(i))==PURL::Nb_FileTypes ) continue;
    _toolType->insertItem(i18n(PURL::TOOL_TYPE_LABELS[i]));
    if ( i==uint(stype) ) _toolType->setCurrentItem(k);
    k++;
  }
  _toolType->setEnabled( _toolType->count()>=2 );
}

void ConfigGenerator::set()
{
  delete _memory;
  const Device::Data *data = Device::lister().data(_deviceChooser->device());
  _memory = data->group().createMemory(*data);
  delete _configEditor;
  _configEditor = Device::groupui(*data).createConfigEditor(*_memory, mainWidget());
  if (_configEditor) {
    _configEditor->show();
    connect(_configEditor, SIGNAL(modified()), SLOT(computeConfig()));
    _configEditor->updateDisplay();
    _hbox->addWidget(_configEditor);
  }
  computeConfig();
}

void ConfigGenerator::computeConfig()
{
  const Tool::Group &group = Tool::lister().group(_configType->currentItem());
  PURL::ToolType stype = PURL::Nb_ToolTypes;
  uint k = 0;
  for (uint i=0; i<PURL::Nb_ToolTypes; i++) {
    if ( group.sourceFileType(PURL::ToolType(i))==PURL::Nb_FileTypes ) continue;
    if ( k==uint(_toolType->currentItem()) ) stype = PURL::ToolType(i);
    k++;
  }
  _warning->setText(QString::null);
  if ( group.isSupported(_memory->device().name()) ) {
    bool ok;
    QString s = group.generateDeviceConfig(stype, *_memory, ok);
    if ( s.isEmpty() ) {
      _text->setFileType(PURL::Nb_FileTypes);
      _text->setText(i18n("Config generation is not supported for the selected toolchain or device."));
    } else {
      _text->setFileType(group.sourceFileType(stype));
      _text->setText(s);
      if ( !ok ) _warning->setText(i18n("Config generation is only partially supported for this device."));
    }
  } else {
   _text->setFileType(PURL::Nb_FileTypes);
    if ( group.supportedDevices().isEmpty() ) _text->setText(i18n("Could not detect supported devices for selected toolchain. Please check installation."));
    else _text->setText(i18n("Device not supported by the selected toolchain."));
  }
}

void ConfigGenerator::slotUser1()
{
  _text->selectAll();
  _text->copy();
  _text->deselect();
}
