/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "editor.h"

#include <qregexp.h>
#include <klocale.h>
#include <kmessagebox.h>

#include "common/gui/purl_gui.h"

const Editor::ActionData Editor::ACTION_DATA[] = {
  { KStdAction::ActionNone, I18N_NOOP("&Split View"), "view_left_right", Qt::CTRL+Qt::SHIFT+Qt::Key_L, false,
    "view_split_view", SLOT(addView()), View, None },
  { KStdAction::ActionNone, I18N_NOOP("&Close Current View"), "view_remove", Qt::CTRL+Qt::SHIFT+Qt::Key_R, false,
    "view_remove_view", SLOT(removeCurrentView()), View, MultipleViews },
  { KStdAction::SelectAll,  0, 0, 0, false, 0, SLOT(selectAll()), Selection, None },
  { KStdAction::Deselect,  0, 0, 0, false, 0, SLOT(deselect()), Selection, SelectionEnabled },
  { KStdAction::Undo,  0, 0, 0, false, 0, SLOT(undo()), UndoRedo, UndoEnabled },
  { KStdAction::Redo,  0, 0, 0, false, 0, SLOT(redo()), UndoRedo, RedoEnabled },
  { KStdAction::Cut,   0, 0, 0, false, 0, SLOT(cut()), Selection, SelectionEnabled | ReadWriteEnabled },
  { KStdAction::Copy,  0, 0, 0, false, 0, SLOT(copy()), Selection, SelectionEnabled },
  { KStdAction::Paste, 0, 0, 0, false, 0, SLOT(paste()), Selection, ReadWriteEnabled }, // #### FIXME
  { KStdAction::Find,  0, 0, 0, false, 0, SLOT(search()), Search, None },
  { KStdAction::FindNext, 0, 0, 0, false, 0, SLOT(findNext()), Search, None }, // #### FIXME
  { KStdAction::FindPrev, 0, 0, 0, false, 0, SLOT(findPrevious()), Search, None }, // #### FIXME
  { KStdAction::Replace,  0, 0, 0, false, 0, SLOT(replace()), Search, ReadWriteEnabled },
  { KStdAction::GotoLine, 0, 0, 0, false, "edit_goto_line", SLOT(gotoLine()), Line, None },
  { KStdAction::Save,   0, 0, 0, false, 0, SLOT(save()), LoadSave, Modified },
  { KStdAction::SaveAs, 0, 0, 0, false, 0, SLOT(saveAs()), LoadSave, None },
  { KStdAction::Print,  0, 0, 0, false, 0, SLOT(printDialog()), Print, None },
  { KStdAction::ActionNone, I18N_NOOP("Show Folding &Markers"), 0, Qt::Key_F9, true,
    "view_folding_markers", SLOT(toggleFoldingMarkers()), Line, FoldingMarkersOn },
  { KStdAction::ActionNone, I18N_NOOP("&Dynamic Word Wrap"), 0, 0, true,
    "view_auto_wrap", SLOT(toggleAutoWrapOn()), Line, AutoWrapOn },
  { KStdAction::ActionNone, I18N_NOOP("Dos End Of Line"), 0, 0, true,
    "view_dos_eol", SLOT(toggleDosEolOn()), Line, DosEolOn | ReadWriteEnabled },
  { KStdAction::ActionNone, I18N_NOOP("Com&ment"), 0, Qt::CTRL+Qt::Key_D, false,
    "edit_comment", SLOT(comment()), Line, ReadWriteEnabled },
  { KStdAction::ActionNone, I18N_NOOP("Uncomme&nt"), 0, Qt::CTRL+Qt::SHIFT+Qt::Key_D, false,
    "edit_uncomment", SLOT(uncomment()), Line, ReadWriteEnabled },
  { KStdAction::ActionNone, I18N_NOOP("&Read Write Mode"), 0, 0, true,
    "view_readwritemode", SLOT(toggleReadWrite()), NoCapability, ReadWriteEnabled },
  { KStdAction::ActionNone, 0, 0, 0, false, 0, 0, NoCapability, None }
};

const char *Editor::actionName(uint i)
{
  if (ACTION_DATA[i].action) return ACTION_DATA[i].action;
  return KStdAction::name(ACTION_DATA[i].stdAction);
}

Editor::Editor(Capabilities capabilities, const QString &title, const QString &tag, QWidget *parent, const char *name)
  : QWidget(parent, name), _capabilities(capabilities), _title(title), _tag(tag)
{}

Editor::Editor(Capabilities capabilities, const PURL::Url &url, QWidget *parent, const char *name)
  : QWidget(parent, name), _capabilities(capabilities), _url(url)
{}

QSizePolicy Editor::sizePolicy() const
{
  return QSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
}

void Editor::setModified(bool m)
{
  setState(Modified, m);
  if (m) emit modified();
}

bool Editor::state(ValueTypes types) const
{
  uint i = 1;
  for (;;) {
    if ( i>=ValueTypes_End ) break;
    if ( (types & ValueType(i)) && !state(ValueType(i)) ) return false;
    i *= 2;
  }
  return true;
}

void Editor::save()
{
  if ( _url.isEmpty() ) {
    saveAs();
    return;
  }
  if ( !save(_url) ) return;
  setModified(false);
  emit guiChanged();
}

void Editor::saveAs()
{
  QString filter = PURL::filter(fileType());
  PURL::Url purl = PURL::getSaveUrl(":save_file_as", filter, this, i18n("Save File"), PURL::AskOverwrite);
  if ( purl.isEmpty() ) return;
  if ( !save(purl) ) return;
  _url = purl;
  setModified(false);
  emit guiChanged();
  return;
}

bool Editor::slotLoad()
{
  bool readOnly = isReadOnly();
  if ( !load() ) return false;
  setModified(false);
  setReadOnly(readOnly);
  statusChanged();
  emit guiChanged();
  return true;
}

QString Editor::filename() const
{
  return (_url.isEmpty() ? "<" + _title + ">" : "\"" + _url.filepath(false) + "\"");
}

bool Editor::checkSaved()
{
  if ( !isModified() ) return true;
  int res = KMessageBox::questionYesNoCancel(this, i18n("File %1 not saved.").arg(filename()),
                                             0, i18n("Save"), i18n("Discard"), i18n("Cancel"));
  switch (res) {
    case KMessageBox::Yes: save(); break;
    case KMessageBox::Cancel: return false;
  }
  return true;
}

bool Editor::reload()
{
  if ( !checkSaved() ) return false;
  return slotLoad();
}
