/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#ifndef EDITOR_H
#define EDITOR_H

#include <qlabel.h>
#include <qlayout.h>
#include <qvaluevector.h>
#include "common/common/qflags.h"
#include <kstdaction.h>
class KPopupMenu;

#include "common/global/purl.h"

class Editor : public QWidget
{
Q_OBJECT
public:
  enum Capability { NoCapability = 0, LoadSave = 1, Selection = 2, Search = 4, Line = 8, Print = 16,
                    View = 32, UndoRedo = 64, CanConfigure = 128, Breakpoints = 256 };
  Q_DECLARE_FLAGS(Capabilities, Capability)
  enum ValueType { None = 0, Modified = 1, MultipleViews = 2, UndoEnabled = 4,
                   RedoEnabled = 8, SelectionEnabled = 16, LineNumbersOn = 32,
                   IconBorderOn = 64, AutoWrapOn = 128, DosEolOn = 256,
                   ReadWriteEnabled = 512, FoldingMarkersOn = 1024, ValueTypes_End };
  Q_DECLARE_FLAGS(ValueTypes, ValueType)
  struct ActionData {
    KStdAction::StdAction stdAction;
    const char *label, *icon;
    int key;
    bool toggle;
    const char *action, *slot;
    Capability capability;
    ValueTypes types;
  };
  static const ActionData ACTION_DATA[];
  static const char *actionName(uint i);

public:
  Editor(Capabilities capabilities, const QString &title, const QString &tag, QWidget *parent, const char *name);
  Editor(Capabilities capabilities, const PURL::Url &url, QWidget *parent, const char *name);

  virtual QSizePolicy sizePolicy() const;
  Capabilities capabilities() const { return _capabilities; }
  virtual bool state(ValueType type) const = 0;
  bool state(ValueTypes types) const;
  virtual void setState(ValueType type, bool on) = 0;
  void toggleState(ValueType type) { setState(type, !state(type)); }
  virtual PURL::FileType fileType() const = 0;
  bool isModified() const { return state(Modified); }
  void setModified(bool modified);
  PURL::Url url() const { return _url; }
  QString name() const { return _title; }
  QString tag() const { return _tag; }
  void setReadOnly(bool readOnly) { setState(ReadWriteEnabled, !readOnly); }
  bool isReadOnly() const { return !state(ReadWriteEnabled); }
  virtual void setCursor(uint line, uint column) = 0;
  virtual void configureEditor() = 0;
  bool checkSaved();
  bool reload();
  virtual void setFocus() = 0;
  virtual bool save(const PURL::Url &url) const = 0;

public slots:
  bool slotLoad();
  void save();
  void saveAs();
  virtual void statusChanged() = 0;
  void toggleReadWrite() { toggleState(ReadWriteEnabled); }
  void toggleLineNumbersOn() { toggleState(LineNumbersOn); }
  void toggleIconBorderOn() { toggleState(IconBorderOn); }
  void toggleAutoWrapOn() { toggleState(AutoWrapOn); }
  void toggleDosEolOn() { toggleState(DosEolOn); }
  void toggleFoldingMarkers() { toggleState(FoldingMarkersOn); }

signals:
  void modified();
  void updateActions();
  void guiChanged();
  void statusTextChanged(const QString &text);
  void dropEventPass(QDropEvent *e);

protected:
  QString filename() const;
  virtual bool load() = 0;

private:
  Capabilities _capabilities;
  QString      _title, _tag;
  PURL::Url    _url;
};
Q_DECLARE_OPERATORS_FOR_FLAGS(Editor::Capabilities)
Q_DECLARE_OPERATORS_FOR_FLAGS(Editor::ValueTypes)

#endif
