/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *   Copyright (C) 2003 Alain Gibaud <alain.gibaud@free.fr>                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "hex_editor.h"

#include <qgroupbox.h>
#include <qhgroupbox.h>
#include <qregexp.h>
#include <qlayout.h>
#include <qscrollview.h>
#include <qstringlist.h>
#include <qlabel.h>
#include <qtimer.h>

#include <klocale.h>
#include <ktempfile.h>

#include "devices/base/device_group.h"
#include "devices/gui/device_group_ui.h"
#include "devices/gui/hex_view.h"
#include "toplevel.h"
#include "common/global/pfile.h"
#include "main_global.h"

HexEditor::HexEditor(const QString &name, QWidget *parent)
  : DeviceEditor(LoadSave, name, QString::null, parent, "hex_editor")
{
  init();
}

HexEditor::HexEditor(const PURL::Url &url, QWidget *parent)
  : DeviceEditor(LoadSave, url, parent, "hex_editor")
{
  init();
}

void HexEditor::init()
{
  _modified = false;
  _memory = 0;
  _originalMemory = 0;
  _top->addStretch(1);
  setReadOnly(false);
}

void HexEditor::clear()
{
  delete _memory;
  _memory = 0;
  delete _originalMemory;
  _originalMemory = 0;
}

QWidget *HexEditor::createView(const Device::Data *data, QWidget *parent)
{
  clear();
  if (data) {
    _originalMemory = data->group().createMemory(*data);
    _memory = data->group().createMemory(*data);
  }
  if ( data==0 ) return new QWidget(parent);
  Device::HexView *hv = Device::groupui(*data).createHexView(*this, parent);
  connect(hv, SIGNAL(modified()), SLOT(slotModified()));
  _dirty = true;
  QTimer::singleShot(0, this, SLOT(simpleLoad()));
  return hv;
}

bool HexEditor::load()
{
  PURL::File file(url(), &Main::compileLog());
  if ( !file.openForRead() ) return false;
  QString error;
  if ( !_hexBuffer.load(*file.qfile(), error) ) {
    MessageBox::sorry(error, Log::Show);
    return false;
  }
  _dirty = true;
  return simpleLoad();
}

bool HexEditor::simpleLoad()
{
  if ( !_dirty ) return true;
  _dirty = false;
  if (_memory) {
    QStringList warnings;
    if ( _memory->fromHexBuffer(_hexBuffer, warnings)!=Device::Memory::NoWarning ) {
      _labelWarning->setText(i18n("<b>Warning:</b> Hex file seems to be incompatible with the selected device %1:<br>%2")
                             .arg(_memory->device().name()).arg(warnings.join("<br>")));
      _labelWarning->show();
    } else _labelWarning->hide();
    display();
  }
  return true;
}

bool HexEditor::state(ValueType type) const
{
  switch (type) {
  case None: return true;
  case Modified:
    return _modified;
  case MultipleViews:
  case UndoEnabled:
  case RedoEnabled:
  case SelectionEnabled:
  case LineNumbersOn:
  case IconBorderOn:
  case AutoWrapOn:
  case FoldingMarkersOn:
  case DosEolOn: break;
  case ReadWriteEnabled: return _readWrite;
  case ValueTypes_End: break;
  }
  Q_ASSERT(false);
  return false;
}

void HexEditor::setState(ValueType type, bool on)
{
  switch (type) {
  case None: return;
  case Modified:
    _modified = on;
    return;
  case MultipleViews:
  case UndoEnabled:
  case RedoEnabled:
  case SelectionEnabled:
  case LineNumbersOn:
  case IconBorderOn:
  case AutoWrapOn:
  case FoldingMarkersOn:
  case DosEolOn: break;
  case ReadWriteEnabled:
    _readWrite = on;
    if (_memory) static_cast<Device::HexView *>(_view)->setReadOnly(!on);
    return;
  case ValueTypes_End: break;
  }
  Q_ASSERT(false);
}

bool HexEditor::save(const PURL::Url &url) const
{
  return save(url, i18n("File URL: \"%1\".").arg(url.pretty()));
}

bool HexEditor::save(const PURL::Url &url, const QString &fileErrorString) const
{
  PURL::File file(url, &Main::compileLog());
  if ( !file.openForWrite() ) return false;
  if ( !_memory->save(*file.qfile(), HexBuffer::IHX32) ) {
    MessageBox::detailedSorry(i18n("Error while writing file \"%1\".").arg(url.pretty()), fileErrorString, Log::Show);
    return false;
  }
  _originalMemory->copyFrom(*_memory);
  return file.close();
}

void HexEditor::display()
{
  _modified = false;
  if (_memory) {
    _originalMemory->copyFrom(*_memory);
    static_cast<Device::HexView *>(_view)->display(_memory);
    static_cast<Device::HexView *>(_view)->setReadOnly(!_readWrite);
    static_cast<Device::HexView *>(_view)->updateDisplay();
  }
  statusChanged();
}

void HexEditor::memoryRead()
{
  display();
  emit guiChanged();
}

void HexEditor::slotModified()
{
  static_cast<Device::HexView *>(_view)->updateDisplay();
  _modified = true;
  statusChanged();
  emit guiChanged();
}

void HexEditor::statusChanged()
{
  QString s;
  if (_memory) {
    uint cs = static_cast<Device::HexView *>(_view)->checksum();
    s = i18n("Checksum: %1").arg(toHexLabel(cs, 4));
  }
  emit statusTextChanged(s);
}
