/***************************************************************************
 *   Copyright (C) 2006 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "object_view.h"

#include <klocale.h>

#include "devices/base/device_group.h"
#include "devices/pic/pic/pic_memory.h"
#include "tools/gputils/gputils.h"
#include "hex_editor.h"
#include "coff/base/text_coff.h"
#include "main_global.h"
#include "tools/base/compile.h"
#include "coff/base/disassembler.h"
#include "common/gui/misc_gui.h"

//-----------------------------------------------------------------------------
Coff::Editor::Editor(const PURL::Url &url, const Device::Data &data, Log::Base &log, QWidget *parent)
  : SimpleTextEditor(Breakpoints, url, parent), _provided(0)
{
  _created = new TextObject(data, url, log, 0);
  setReadOnly(true);
  setState(AutoWrapOn, false);
  setView(&Main::compileLog());
}

Coff::Editor::Editor(const PURL::Url &url, const TextObject &object, QWidget *parent)
  : SimpleTextEditor(Breakpoints, url, parent), _created(0), _provided(&object)
{
  setReadOnly(true);
  setState(AutoWrapOn, false);
  setView(&Main::compileLog());
}

Coff::Editor::~Editor()
{
  delete _created;
}

bool Coff::Editor::load()
{
  if (_provided) return setText(_provided->text());
  return setText(_created->text());
}

//-----------------------------------------------------------------------------
DisassemblyEditor::DisassemblyEditor(const PURL::Url &url, const PURL::Url &hexUrl,
                                     const Device::Data &data, QWidget *parent)
  : SimpleTextEditor(Breakpoints, url, parent), _data(data), _hexUrl(hexUrl), _editor(0)
{
  setReadOnly(true);
  setView(&Main::compileLog());
}

DisassemblyEditor::DisassemblyEditor(const PURL::Url &url, const HexEditor &e,
                                     const Device::Data &data, QWidget *parent)
  : SimpleTextEditor(Breakpoints, url, parent), _data(data), _editor(&e)
{
  setReadOnly(true);
  setView(&Main::compileLog());
}

bool DisassemblyEditor::load()
{
  if ( Main::toolGroup().name()!="sdcc" && Main::toolGroup().name()!="gputils" ) {
    MessageBox::sorry(i18n("Disassembly not supported for the selected toolchain."), Log::Show);
    return false;
  }
  Pic::Architecture arch = (_data.group().name()=="pic" ? static_cast<const Pic::Data &>(_data).architecture() : Pic::Nb_Architectures);
  if ( arch==Pic::Nb_Architectures ) {
    MessageBox::sorry(i18n("Disassembly not supported for the selected device."), Log::Show);
    return false;
  }

  Device::Memory *memory = 0;
  if ( _editor==0 ) {
    log(Log::Information, i18n("Disassembling hex file: %1").arg(_hexUrl.pretty()));
    PURL::File file(_hexUrl, &Main::compileLog());
    if ( !file.openForRead() ) return false;
    memory = _data.group().createMemory(_data);
    QString error;
    Device::Memory::WarningTypes warningTypes;
    QStringList warnings;
    if ( !memory->load(*file.qfile(), error, warningTypes, warnings) ) {
      delete memory;
      log(Log::Error, error);
      return false;
    }
    for (uint i=0; i<warnings.count(); i++) log(Log::Warning, warnings[i]);
  } else {
    log(Log::Information, i18n("Disassembling content of hex editor."));
    memory = const_cast<Device::Memory *>(_editor->memory());
  }
  QString s = GPUtils::disassemble(static_cast<const Pic::Memory &>(*memory));
  if ( _editor==0 ) delete memory;
  return setText(s);
}
