/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *   Copyright (C) 2003-2004 Alain Gibaud <alain.gibaud@free.fr>           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "direct_config_widget.h"

#include <qhbox.h>
#include <qvbox.h>
#include <qlabel.h>
#include <qlayout.h>
#include <qgroupbox.h>
#include <qtooltip.h>
#include <qspinbox.h>

#include <klocale.h>
#include <kiconloader.h>
#include <kmessagebox.h>

#include "progs/direct/base/direct_prog_config.h"
#include "progs/base/prog_config.h"
#include "devices/base/device_group.h"
#include "progs/direct/base/direct_prog.h"
#include "progs/list/prog_list.h"

//-----------------------------------------------------------------------------
::Programmer::ConfigWidget *Direct::GroupUI::createConfigWidget(QWidget *parent) const
{
  return new ConfigWidget(static_cast<const ::Programmer::Group &>(group()), parent);
}

//-----------------------------------------------------------------------------
const char * const INV_PIN_LABEL = I18N_NOOP("Check this option if your hardware uses negative logic for this pin.");
const char * const DELAY_LABEL = I18N_NOOP("Some programming cards need low clock rate:\nadding delay to clock pulses might help.");

Direct::HardwareConfigWidget::HardwareConfigWidget(const ::Programmer::Group &group, QWidget *parent, bool edit)
  : QFrame(parent, "hardware_config"), _edit(edit), _connected(false)
{
  _base = new PicBase(group, 0); // dummy
  _hardware = 0;

  QHBoxLayout *top = new QHBoxLayout(this, 0, 10);
  QVBoxLayout *vbox = new QVBoxLayout(top);

  // pins assignment
  QGroupBox *groupb = new QGroupBox(1, Horizontal, i18n("Pin assignment"), this);
  vbox->addWidget(groupb);
  QWidget *w = new QWidget(groupb);
  QGridLayout *grid = new QGridLayout(w, 1, 1, 0, 10);
  if (edit) grid->setColStretch(5, 1);
  for (uint i=0; i<Nb_PinTypes; i++) {
    QLabel *label = new QLabel(i18n(PIN_DATA[i].label), w);
    QToolTip::add(label, PIN_DATA[i].comment);
    grid->addWidget(label, i, 0);
    _combos[i] = new QComboBox(w);
    _combos[i]->setEnabled(edit);
    connect(_combos[i], SIGNAL(activated(int)), SLOT(slotPinChanged()));
    QToolTip::add(_combos[i], PIN_DATA[i].comment);
    grid->addWidget(_combos[i], i, 1);
    _invcbs[i] = new QCheckBox(i18n("Inverted"), w);
    _invcbs[i]->setEnabled(edit);
    QToolTip::add(_invcbs[i], i18n(INV_PIN_LABEL));
    grid->addWidget(_invcbs[i], i, 2);
    if (edit) {
      _testcbs[i] = new QCheckBox(i18n("on"), w);
      QToolTip::add(_testcbs[i], PIN_DATA[i].testComment);
      connect(_testcbs[i], SIGNAL(clicked()), SLOT(slotTestPin()));
      grid->addWidget(_testcbs[i], i, 3);
      _testLabels[i] = new QLabel(w);
      QToolTip::add(_testcbs[i], PIN_DATA[i].testComment);
      grid->addWidget(_testLabels[i], i, 4);
      updateTestStatus(PinType(i), false);
    } else {
      _testcbs[i] = 0;
      _testLabels[i] = 0;
    }
  }

  QHBoxLayout *hbox = new QHBoxLayout(vbox);
  QLabel *label = new QLabel(i18n("Clock delay"), this);
  QToolTip::add(label, i18n(DELAY_LABEL));
  hbox->addWidget(label);
  _delay = new KIntNumInput(0, Horizontal, this);
  _delay->setRange(0, 50, 5);
  _delay->setEnabled(edit);
  QToolTip::add(_delay, i18n(DELAY_LABEL));
  hbox->addWidget(_delay);

  if (edit) {
    QVBoxLayout *vbox = new QVBoxLayout(top);
    top->setStretchFactor(vbox, 1);
    _sendBitsButton = new QPushButton(i18n("Send 0xA55A"), this);
    _sendBitsButton->setToggleButton(true);
    connect(_sendBitsButton, SIGNAL(clicked()), SLOT(sendBits()));
    QToolTip::add(_sendBitsButton, i18n("Continuously send 0xA55A on \"Data out\" pin."));
    vbox->addWidget(_sendBitsButton);
    vbox->addStretch(1);
  } else _sendBitsButton = 0;

  // timer for sending bits
  _timerSendBits = new QTimer(this);
  connect(_timerSendBits, SIGNAL(timeout()), SLOT(slotSendBits()));

  // timer for automatically polling DataOut pin
  _timerPollDataOut = new QTimer(this);
  connect(_timerPollDataOut, SIGNAL(timeout()), SLOT(updateDataIn()));
}

Direct::HardwareConfigWidget::~HardwareConfigWidget()
{
  delete _hardware;
  delete _base;
}

void Direct::HardwareConfigWidget::sendBits()
{
  if ( _timerSendBits->isActive() ) {
    _sendBitsButton->setText(i18n("Send 0xA55A"));
    _timerSendBits->stop();
    updateTestPin(DataOut);
  } else {
    _sendBitsButton->setText(i18n("Stop"));
    _timerSendBits->start(1);
  }
}

uint Direct::HardwareConfigWidget::pin(PinType ptype) const
{
  return _hardware->pinForIndex(PIN_DATA[ptype].dir, _combos[ptype]->currentItem());
}

void Direct::HardwareConfigWidget::slotPinChanged()
{
  for (uint i = 0; i<Nb_PinTypes; i++) {
    if ( sender()!=_combos[i] ) continue;
    updatePin(PinType(i));
  }
}

void Direct::HardwareConfigWidget::updatePin(PinType ptype)
{
  bool ground = _hardware->isGroundPin(pin(ptype));
  _invcbs[ptype]->setEnabled(_edit);
  if (ground) _invcbs[ptype]->hide();
  else _invcbs[ptype]->show();
  if (_edit) {
    _testcbs[ptype]->setEnabled(PIN_DATA[ptype].dir==Port::Out && _connected);
    _testLabels[ptype]->setEnabled(_connected);
    if (ground) {
      _testcbs[ptype]->hide();
      _testLabels[ptype]->hide();
    } else {
      _testcbs[ptype]->show();
      _testLabels[ptype]->show();
    }
  }
}

void Direct::HardwareConfigWidget::slotTestPin()
{
  for (uint i = 0; i<Nb_PinTypes; i++) {
    if ( sender()!=_testcbs[i] ) continue;
    updateTestPin(PinType(i));
    break;
  }
}

void Direct::HardwareConfigWidget::updateTestPin(PinType ptype)
{
  Q_ASSERT( _connected && ptype!=DataIn );
  bool on = _testcbs[ptype]->isChecked();
  _hardware->setPin(ptype, on);
  updateTestStatus(ptype, on);
  if ( ptype==Vpp ) updateDataIn();
}

void Direct::HardwareConfigWidget::updateTestStatus(PinType ptype, bool on)
{
  if (on) _testLabels[ptype]->setText(i18n(PIN_DATA[ptype].onLabel));
  else _testLabels[ptype]->setText(i18n(PIN_DATA[ptype].offLabel));
}

void Direct::HardwareConfigWidget::updateDataIn()
{
  bool on = _hardware->readBit();
  updateTestStatus(DataIn, on);
  _testcbs[DataIn]->setChecked(on);
}

void Direct::HardwareConfigWidget::sendBits(uint d, int nbb)
{
  Q_ASSERT(_connected);
  _hardware->setWrite();
  for (; nbb; --nbb) {
    _hardware->setPin(Clock, High);
    if ( d & 0x01 ) _hardware->setPin(DataOut, High);
    else _hardware->setPin(DataOut, Low);
    _hardware->setPin(Clock, Low);
    d >>= 1;       // Move the data over 1 bit
  }
  _hardware->setPin(DataOut, Low);
  _hardware->setRead();
}

void Direct::HardwareConfigWidget::slotSendBits()
{
    sendBits(0xA55A, 16);
}

bool Direct::HardwareConfigWidget::set(const Port::Description &pd, const Hardware::HData &data)
{
  // connect port
  _timerPollDataOut->stop();
  if ( _timerSendBits->isActive() ) sendBits(); // stop
  delete _hardware;
  if ( pd.type==Port::SerialType ) _hardware = new SerialHardware(*_base, pd.device, data);
  else _hardware = new ParallelHardware(*_base, pd.device, data);
  bool ok = _hardware->connectHardware();
  if ( !_edit) _hardware->disconnectHardware();
  else _connected = ok;

  // update GUI
  if (_edit) {
    for (uint i=0; i<Nb_PinTypes; i++) {
      _testcbs[i]->setEnabled(_connected);
      updateTestStatus(PinType(i), false);
    }
    if ( _connected ) _timerPollDataOut->start(100);
    _sendBitsButton->setEnabled(_connected);
  }

  // update pins
  for (uint i=0; i<Nb_PinTypes; i++) {
    _combos[i]->clear();
    Port::IODir dir = PIN_DATA[i].dir;
    for (uint k=0; k<_hardware->nbPins(dir); k++)
      _combos[i]->insertItem(_hardware->pinLabelForIndex(dir, k));
    if (PIN_DATA[i].canBeGround) _combos[i]->insertItem("GND");
    _combos[i]->setCurrentItem(_hardware->indexForPin(dir, data.pins[i]));
    _invcbs[i]->setChecked(data.pins[i]<0);
    updatePin(PinType(i));
  }
  _delay->setValue(data.clockDelay);

  return ok;
}

Direct::Hardware::HData Direct::HardwareConfigWidget::data() const
{
  Hardware::HData data;
  for (uint i=0; i<Nb_PinTypes; i++) {
    data.pins[i] = pin(PinType(i));
    if ( _invcbs[i]->isChecked() ) data.pins[i] = -data.pins[i];
  }
  data.clockDelay = _delay->value();
  return data;
}

//-----------------------------------------------------------------------------
Direct::HardwareConfigDialog::HardwareConfigDialog(const ::Programmer::Group &group,
             QWidget *parent, const QString &name,
             const Port::Description &pd, const Hardware::HData &data)
  : KDialogBase(Plain, i18n("Edit and test hardware"), Ok|Cancel, Cancel,
                parent, "hardware_config_dialog", true, true),
    _savedName(name), _oldData(data)
{
  setButtonOK(i18n("Save"));
  setButtonCancel(i18n("Close"));

  QGridLayout *grid = new QGridLayout(plainPage(), 1, 1, 0, 10);
  grid->setColStretch(2, 1);

  QLabel *label = new QLabel(i18n("Hardware name:"), plainPage());
  grid->addWidget(label, 0, 0);
  _name = new QLineEdit(name, plainPage());
  grid->addWidget(_name, 0, 1);

  label = new QLabel(i18n("%1 at %2:").arg(i18n(Port::DATA[pd.type].label)).arg(pd.device), plainPage());
  grid->addWidget(label, 1, 0);
  label = new QLabel(plainPage());
  grid->addWidget(label, 1, 1);

  _hc = new HardwareConfigWidget(group, plainPage(), true);
  grid->addMultiCellWidget(_hc, 2,2, 0,2);

  grid->setRowStretch(3, 1);

  bool ok = _hc->set(pd, data);
  label->setText(ok ? i18n("Connected") : i18n("Not Connected"));
}

void Direct::HardwareConfigDialog::slotOk()
{
  if ( _name->text().isEmpty() ) {
    KMessageBox::sorry(this, i18n("Could not save configuration: hardware name is empty."));
    return;
  }
  Config config;
  if ( config.isStandardHardware(_name->text()) ) {
    KMessageBox::sorry(this, i18n("The hardware name is already used for a standard hardware."));
    return;
  }
  QStringList names = config.hardwareNames(Port::Nb_Types); // all hardwares
  if ( names.contains(_name->text()) ) {
    int res = KMessageBox::warningContinueCancel(this, i18n("Do you want to overwrite this custom hardware \"%1\"?").arg(_name->text()),
                                                 QString::null, KStdGuiItem::save());
    if ( res==KMessageBox::Cancel ) return;
  }
  HardwareData hd;
  hd.portType = _hc->portDescription().type;
  hd.data = _hc->data();
  config.writeCustomHardware(_name->text(), hd);
  _savedName = _name->text();
  _oldData = hd.data;
  KDialogBase::accept();
}

void Direct::HardwareConfigDialog::slotCancel()
{
  if ( _oldData!=_hc->data() ) {
    int res = KMessageBox::warningContinueCancel(this, i18n("Closing will discard changes you have made. Close anyway?").arg(_name->text()),
                                                 QString::null, KStdGuiItem::close());
    if ( res==KMessageBox::Cancel ) return;
    KDialogBase::reject();
  }
  KDialogBase::accept();
}

//-----------------------------------------------------------------------------
Direct::ConfigWidget::ConfigWidget(const ::Programmer::Group &group, QWidget *parent)
  : ::Programmer::ConfigWidget(group, parent, "direct_config_widget")
{
// programmer combo
  uint row = _topLayout->numRows();
  _configCombo = new QComboBox(this);
  connect(_configCombo, SIGNAL(activated(int)), SLOT(configChanged(int)));
  _topLayout->addWidget(_configCombo, row, 0);
  _comment = new QLabel(this);
  _topLayout->addWidget(_comment, row, 1);
  row++;

// hardware config
  QHBoxLayout *hbox = new QHBoxLayout(10);
  _topLayout->addMultiCellLayout(hbox, row,row, 0,1);
  _hc = new HardwareConfigWidget(group, this, false);
  hbox->addWidget(_hc);

// buttons
  QVBoxLayout *vbox = new QVBoxLayout(hbox);
  QPushButton *button = new QPushButton(i18n("Edit/Test..."), this);
  connect(button, SIGNAL(clicked()), SLOT(editClicked()));
  vbox->addWidget(button);
  _deleteButton = new QPushButton(i18n("Delete"), this);
  connect(_deleteButton, SIGNAL(clicked()), SLOT(deleteClicked()));
  vbox->addWidget(_deleteButton);
  vbox->addStretch(1);
}

void Direct::ConfigWidget::saveConfig()
{
  ::Programmer::ConfigWidget::saveConfig();
  Config config;
  config.writeCurrentHardware(_hc->portDescription().type, _names[_configCombo->currentItem()]);
}

void Direct::ConfigWidget::configChanged(int i)
{
  set(_hc->portDescription(), i);
}

bool Direct::ConfigWidget::set(const Port::Description &pd, uint i)
{
  Config config;
  bool ok = _hc->set(pd, config.hardware(_names[i]).data);
  const StandardHardwareData *shd = config.standardHardwareData(_names[i]);
  QString s = (shd && shd->comment ? i18n(shd->comment) : QString::null);
  _comment->setText(s);
  _deleteButton->setEnabled(shd==0);
  return ok;
}

bool Direct::ConfigWidget::setPort(const Port::Description &pd)
{
  updateList(pd.type);
  Config config;
  int i = _names.findIndex(config.currentHardware(pd.type));
  if ( i!=-1 ) _configCombo->setCurrentItem(i);
  return set(pd, _configCombo->currentItem());
}

void Direct::ConfigWidget::updateList(Port::Type type)
{
  _configCombo->clear();
  Config config;
  _names = config.hardwareNames(type);
  for (uint i=0; i<_names.count(); i++) {
    const StandardHardwareData *shd = config.standardHardwareData(_names[i]);
    QString s = (shd ? i18n(shd->label) : i18n("%1 <custom>").arg(_names[i]));
    _configCombo->insertItem(s);
  }
}

void Direct::ConfigWidget::editClicked()
{
  QString name = _names[_configCombo->currentItem()];
  Config config;
  if ( config.isStandardHardware(name) ) name = QString::null;
  Port::Description pd = _hc->portDescription();
  HardwareConfigDialog *hcd =
    new HardwareConfigDialog(group(), this, name, pd, _hc->data());
  hcd->exec();
  updateList(pd.type);
  if ( !hcd->savedName().isEmpty() )
    _configCombo->setCurrentItem(_names.findIndex(hcd->savedName()));
  delete hcd;
  configChanged(_configCombo->currentItem());
}

void Direct::ConfigWidget::deleteClicked()
{
  QString name = _names[_configCombo->currentItem()];
  int res = KMessageBox::warningContinueCancel(this, i18n("Do you want to delete custom hardware \"%1\"?").arg(name),
                                               QString::null, KStdGuiItem::del());
  if ( res==KMessageBox::Cancel ) return;
  Config config;
  config.deleteCustomHardware(name);
  updateList(_hc->portDescription().type);
  configChanged(_configCombo->currentItem());
}
