/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "icd2_debug.h"

#include <qdir.h>

#include "progs/base/prog_config.h"
#include "devices/pic/pic/pic_debug.h"
#include "devices/pic/base/pic_register.h"
#include "devices/pic/pic/pic_group.h"
#include "icd2_data.h"

//----------------------------------------------------------------------------
uint Icd2::P16FDebuggerSpecific::addressRegister(uint address) const
{
  QString name = device()->registersData().sfrNames[address];
  if ( name=="FSR" ) return 0x1BE;
  if ( name=="STATUS" ) return 0x1F0;
  return address;
}

bool Icd2::P16FDebuggerSpecific::reset()
{
  if ( !hardware()->setTargetReset(Pic::ResetHeld) ) return false;
  return init();
}

bool Icd2::P16FDebuggerSpecific::setBreakpoint(uint address)
{
  if ( address==Device::INVALID ) address = 0x1FFF; // outside memory range
  uint value = address | 0x8000; // read-only INBUG bit
  return hardware()->writeRegister(0x18E, value, 2);
}

bool Icd2::P16FDebuggerSpecific::readBreakpoint(uint &value)
{
  if ( !hardware()->readRegister(0x18E, value, 2) ) return false;
  value &= 0x1FFF;
  return true;
}

bool Icd2::P16FDebuggerSpecific::init()
{
  Device::Array saved;
  if ( !hardware()->readMemory(Pic::MemoryCode, 0, 1, saved, 0) ) return false; // save instruction
  // #### next check if for some devices only...
  if ( saved!=deviceSpecific()->nopInstruction() ) log(Log::Warning, i18n("  According to ICD2 manual, instruction at address 0x0 should be \"nop\"."));

  // this seems to be needed...
  if ( !hardware()->setTargetReset(Pic::ResetReleased) ) return false;
  Pic::TargetMode mode;
  if ( !programmer().getTargetMode(mode) ) return false;
  if ( !hardware()->setTargetReset(Pic::ResetHeld) ) return false;

  double vdd;
  if ( !hardware()->readVoltage(Pic::TargetVdd, vdd) ) return false;
  log(Log::NormalDebug, QString("  Target Vdd: %1 V").arg(vdd));

  if ( !hardware()->setWriteMode(Pic::EraseWriteMode) ) return false;
  log(Log::NormalDebug, "  Try to run target with loop at reset vector.");
  Device::Array data = deviceSpecific()->gotoInstruction(0x0000); // loop at reset vector
  if ( !hardware()->writeMemory(Pic::MemoryCode, data, 0, data.count()) ) return false;
  if ( !hardware()->setTargetReset(Pic::ResetReleased) ) return false; // run device
  if ( !base().waitForTargetMode(Pic::TargetRunning) ) return false;
  if ( !hardware()->haltRun() ) return false;
  if ( !base().waitForTargetMode(Pic::TargetStopped) ) return false;
  if ( !programmer().readDebugExecutiveVersion() ) return false;

  log(Log::NormalDebug, "  Set breakpoint at reset vector and run.");
  if ( !setBreakpoint(0x0000) ) return false;
  if ( !hardware()->setTargetReset(Pic::ResetHeld) ) return false;
  if ( !base().waitForTargetMode(Pic::TargetInProgramming) ) return false;
  if ( !hardware()->writeMemory(Pic::MemoryCode, saved, 0, saved.count()) ) return false; // restore instruction
  if ( !hardware()->setTargetReset(Pic::ResetReleased) ) return false; // run device
  if ( !base().waitForTargetMode(Pic::TargetStopped) ) return false;
  uint value;
  if ( !readBreakpoint(value) ) return false;
  uint expectedPC = 0x0001;
  if ( value==0x0002) {
    expectedPC = 0x0002;
    log(Log::Information, i18n("Detected custom ICD2"));
  }
  if ( value!=expectedPC ) {
    log(Log::Error, i18n("  PC is not at address %1 (%2)").arg(toHexLabel(expectedPC, 4)).arg(toHexLabel(value, 4)));
    return false;
  }
  if ( !setBreakpoint(0x0000) ) return false;

  if ( !hardware()->setWriteMode(Pic::WriteOnlyMode) ) return false;
  if ( !base().update() ) return false;
  if ( base().pc()!=expectedPC ) {
    log(Log::Error, i18n("  PC is not at address %1 (%2)").arg(toHexLabel(expectedPC, 4)).arg(toHexLabel(base().pc(), 4)));
    return false;
  }
  return true;
}

//----------------------------------------------------------------------------
uint Icd2::P18FDebuggerSpecific::reservedRegisterOffset() const
{
  return device()->registersData().lastUsedBank() << 8;
}

uint Icd2::P18FDebuggerSpecific::addressWREG()const
{
  return reservedRegisterOffset() | 0x0FF;
}

uint Icd2::P18FDebuggerSpecific::addressRegister(uint address) const
{
  QString name = device()->registersData().sfrNames[address];
  if ( name=="PCLATU" ) return reservedRegisterOffset() | 0x0F4;
  if ( name=="PCLATH" ) return reservedRegisterOffset() | 0x0F5;
  if ( name=="FSR0H" ) return reservedRegisterOffset() | 0x0FB;
  if ( name=="FSR0L" ) return reservedRegisterOffset() | 0x0FC;
  if ( name=="BSR" ) return reservedRegisterOffset() | 0x0FD;
  if ( name=="STATUS" ) return reservedRegisterOffset() | 0x0FE;
  return address;
}

bool Icd2::P18FDebuggerSpecific::setBreakpoint(uint address)
{
  uint value = (address==Device::INVALID ? 0x0FFFFF00 : address << 15); // ??
  return hardware()->writeRegister(0xFB6, value, 4);
}

bool Icd2::P18FDebuggerSpecific::readBreakpoint(uint &value)
{
  return hardware()->readRegister(0xFB6, value, 4) >> 15;
}

bool Icd2::P18FDebuggerSpecific::reset()
{
  if ( !hardware()->writeRegister(0xFB5, 0x00, 1) ) return false; // #### ??
  if ( !hardware()->writeRegister(0xFB5, 0x01, 1) ) return false; // #### ??
  if ( !hardware()->command("2D", 0) ) return false; // reset
  if ( !hardware()->writeRegister(0xFB5, 0x00, 1) ) return false; // #### ??
  if ( !base().update() ) return false;
  uint expectedPC = 0x0000;
  if ( base().pc()==0x0001) {
    expectedPC = 0x0001;
    log(Log::Information, i18n("Detected custom ICD2"));
  }
  if ( base().pc()!=expectedPC ) {
    log(Log::Error, i18n("  PC is not at address %1 (%2)").arg(toHexLabel(expectedPC, 4)).arg(toHexLabel(base().pc(), 4)));
    return false;
  }
  return true;
}

bool Icd2::P18FDebuggerSpecific::init()
{
  if ( !hardware()->setTargetReset(Pic::ResetReleased) ) return false;
  if ( !base().waitForTargetMode(Pic::TargetStopped) ) return false;
  return reset();
}

//-----------------------------------------------------------------------------
Icd2::DebuggerSpecific *Icd2::Debugger::specific() { return static_cast<Icd2::DebuggerSpecific *>(_specific); }

bool Icd2::Debugger::waitForTargetMode(Pic::TargetMode mode)
{
  Pic::TargetMode rmode;
  for (uint i=0; i<20; i++) {
    if ( !programmer().getTargetMode(rmode) ) return false;
    if ( rmode==mode ) return true;
    Port::msleep(200);
  }
  log(Log::Error, QString("Timeout waiting for mode: %1 (target is in mode: %2).")
                  .arg(i18n(Pic::TARGET_MODE_LABELS[mode])).arg(i18n(Pic::TARGET_MODE_LABELS[rmode])));
  return false;
}

bool Icd2::Debugger::internalInit()
{
  return specific()->init();
}

bool Icd2::Debugger::updateState()
{
  Pic::TargetMode mode;
  if ( !programmer().getTargetMode(mode) ) return false;
  switch (mode) {
    case Pic::TargetStopped: _programmer.setState(::Programmer::Halted); break;
    case Pic::TargetRunning: _programmer.setState(::Programmer::Running); break;
    case Pic::TargetInProgramming: _programmer.setState(::Programmer::Stopped); break;
    case Pic::Nb_TargetModes: Q_ASSERT(false); break;
  }
  return true;
}

bool Icd2::Debugger::setBreakpoints(const QValueList<uint> &addresses)
{
  if ( addresses.count()==0 ) return specific()->setBreakpoint(Device::INVALID);
  for (uint i=0; i<uint(addresses.count()); i++) {
    log(Log::NormalDebug, QString("Set breakpoint at %1").arg(toHexLabel(addresses[i], device()->nbCharsAddress())));
    if ( !specific()->setBreakpoint(addresses[i]) ) return false;
  }
  return true;
}

bool Icd2::Debugger::internalRun()
{
  return hardware()->resumeRun();
}

bool Icd2::Debugger::hardHalt()
{
  log(Log::Warning, i18n("Failed to halt target: try a reset."));
  return reset();
}

bool Icd2::Debugger::softHalt(bool &success)
{
  if ( !hardware()->haltRun() ) return false;
  success = waitForTargetMode(Pic::TargetStopped);
  return true;
}

bool Icd2::Debugger::internalReset()
{
  return specific()->reset();
}

bool Icd2::Debugger::internalStep()
{
  return hardware()->step();
}

bool Icd2::Debugger::readRegister(const Register::TypeData &data, uint &value)
{
  if ( data.type()==Register::Special ) {
    if ( data.name()=="WREG" ) return hardware()->readRegister(specific()->addressWREG(), value, 1);
    if ( data.name()=="PC" ) { value = hardware()->getProgramCounter() & specific()->maskPC(); return !hasError(); }
    Q_ASSERT(false);
    return true;
  }
  QString name = device()->registersData().sfrNames[data.address()];
  if ( name=="WREG" ) return hardware()->readRegister(specific()->addressWREG(), value, 1);
  if ( name=="PCL" ) { value = (hardware()->getProgramCounter() & specific()->maskPC()) && 0xFF; return !hasError(); }
  if ( name=="PCLATH" ) { value = (hardware()->getProgramCounter() & specific()->maskPC()) >> 8; return !hasError(); }
  return hardware()->readRegister(specific()->addressRegister(data.address()), value, 1);
}

bool Icd2::Debugger::writeRegister(const Register::TypeData &data, uint value)
{
  if ( data.type()==Register::Special ) {
    if ( data.name()=="WREG" ) return hardware()->writeRegister(specific()->addressWREG(), value, 1);
    Q_ASSERT(false);
    return true;
  }
  QString name = device()->registersData().sfrNames[data.address()];
  if ( name=="WREG" ) return hardware()->writeRegister(specific()->addressWREG(), value, 1);
  return hardware()->writeRegister(specific()->addressRegister(data.address()), value, 1);
}

//-----------------------------------------------------------------------------
Icd2::DebugProgrammer::DebugProgrammer(const ::Programmer::Group &group, const Pic::Data *data)
  : Icd2::ProgrammerBase(group, data, "icd2_programmer")
{}

void Icd2::DebugProgrammer::clear()
{
  Icd2::ProgrammerBase::clear();
  _debugExecutiveVersion.clear();
}

bool Icd2::DebugProgrammer::internalSetupHardware()
{
  if ( _specific==0 ) return true;

  // find debug executive file
  ::Programmer::GroupConfig config(group());
  QDir dir(config.firmwareDirectory());
  if ( !dir.exists() ) {
    log(Log::Error, i18n("Firmware directory not configured."));
    return false;
  }
  QString filename;
  if ( device()->architecture()==Pic::Arch_18X ) {
    uint i = device()->registersData().lastUsedBank();
    filename = QString("de18F_BANK%1.hex").arg(QString(toString(Number::Dec, i, 2)));
  } else filename = QString("de%1.hex").arg(FAMILY_DATA[family(device()->name())].debugExec);
  QString filepath = dir.filePath(filename);
  log(Log::NormalDebug, QString("  Debug executive file: %1").arg(filepath));
  QFile file(filepath);
  if ( !file.exists() ) {
    log(Log::Error, i18n("Could not find debug executive file \"%1\".").arg(file.name()));
    return false;
  }
  // upload hex file
  if ( !file.open(IO_ReadOnly) ) {
    log(Log::Error, i18n("Could not open firmware file \"%1\".").arg(file.name()));
    return false;
  }
  QString error;
  QStringList warnings;
  Pic::Memory::WarningTypes warningTypes;
  Pic::Memory memory(*device());
  if ( !memory.load(file, error, warningTypes, warnings) ) {
    log(Log::Error, i18n("Could not read debug executive file \"%1\" (%2).").arg(file.name()).arg(error));
    return false;
  }
  if ( warningTypes!=Pic::Memory::NoWarning ) {
    log(Log::Error, i18n("Debug executive file seems incompatible with current device:\n%1.").arg(warnings.join("\n")));
    return false;
  }
  _deArray = memory.arrayForWriting(Pic::MemoryCode);
  _deEnd = specific()->findNonMaskEnd(Pic::MemoryCode, _deArray);
  if ( device()->architecture()==Pic::Arch_18X ) {
    _deStart = device()->nbWords(Pic::MemoryCode) - _deEnd - 1;
    for (uint i=0; i<=_deEnd; i++) {
      uint v = memory.word(Pic::MemoryCode, i);
      memory.setWord(Pic::MemoryCode, i, Device::INVALID);
      memory.setWord(Pic::MemoryCode, _deStart+i, v);
    }
    _deEnd = device()->nbWords(Pic::MemoryCode) - 1 ;
    _deArray = memory.arrayForWriting(Pic::MemoryCode);
  } else _deStart = specific()->findNonMaskStart(Pic::MemoryCode, _deArray);
  log(Log::ExtraDebug, QString("debug executive: \"%1\" %2:%3").arg(file.name()).arg(toHexLabel(_deStart, 4)).arg(toHexLabel(_deEnd, 4)));

  return Icd2::ProgrammerBase::internalSetupHardware();
}

Pic::Memory Icd2::DebugProgrammer::toDebugMemory(const Pic::Memory &mem, bool withDebugExecutive)
{
  Pic::Memory memory = mem;
  memory.setDebugOn(true);
  if ( memory.hasWatchdogTimerOn() ) {
    log(Log::Warning, i18n("Disabling watchdog timer for debugging"));
    memory.setWatchdogTimerOn(false);
  }
  if ( memory.isProtected(Pic::Config::ProgramProtected, Pic::Config::CodeBlock) ) {
    log(Log::Warning, i18n("Disabling code program protection for debugging"));
    memory.setProtection(false, Pic::Config::ProgramProtected, Pic::Config::CodeBlock);
  }
  if ( memory.isProtected(Pic::Config::WriteProtected, Pic::Config::CodeBlock) ) {
    log(Log::Warning, i18n("Disabling code write protection for debugging"));
    memory.setProtection(false, Pic::Config::WriteProtected, Pic::Config::CodeBlock);
  }
  if ( memory.isProtected(Pic::Config::ReadProtected, Pic::Config::CodeBlock) ) {
    log(Log::Warning, i18n("Disabling code read protection for debugging"));
    memory.setProtection(false, Pic::Config::ReadProtected, Pic::Config::CodeBlock);
  }
  Device::Array data = static_cast<Debugger::PicBase *>(_debugger)->deviceSpecific()->gotoInstruction(_deStart);
  for (uint i=0; i<data.count(); i++) memory.setWord(Pic::MemoryDebugVector, i, data[i]);
  if ( device()->architecture()==Pic::Arch_18X ) {
    memory.setWord(Pic::MemoryDebugVector, data.count(), 0xFF00); // ??
    for (uint i=data.count()+1; i<device()->nbWords(Pic::MemoryDebugVector); i++)
      memory.setWord(Pic::MemoryDebugVector, i, 0x0000); // seems to be needed for some chips ?? (should be harmless)
  }
  if (withDebugExecutive) {
    bool ok = true;
    for (uint i=_deStart; i<=_deEnd; i++) {
      if ( memory.word(Pic::MemoryCode, i)!=Device::INVALID ) ok = false;
      memory.setWord(Pic::MemoryCode, i, _deArray[i]);
    }
    if ( !ok ) log(Log::Warning, i18n("Memory area for debug executive was not empty. Overwrite it and continue anyway..."));
  }
  return memory;
}

bool Icd2::DebugProgrammer::writeDebugExecutive()
{
  log(Log::Information, i18n("  Write debug executive"));
  uint nbWords = _deEnd - _deStart + 1;
  if ( !hardware()->writeMemory(Pic::MemoryCode, _deArray, _deStart, nbWords) ) return false;
  log(Log::Information, i18n("  Verify debug executive"));
  Device::Array data;
  if ( !hardware()->readMemory(Pic::MemoryCode, _deStart, nbWords, data, 0) ) return false;
  for (uint i=0; i<nbWords; i++) {
    if ( _deArray[_deStart+i]==data[i] ) continue;
    uint inc = device()->addressIncrement(Pic::MemoryCode);
    uint address = device()->range(Pic::MemoryCode).start + inc * (_deStart + i);
    log(Log::Error, i18n("Device memory doesn't match debug executive (at address %1: reading %2 and expecting %3).")
                    .arg(toHexLabel(address, device()->nbCharsAddress()))
                    .arg(toHexLabel(data[i], device()->nbCharsWord(Pic::MemoryCode)))
                    .arg(toHexLabel(_deArray[_deStart+i], device()->nbCharsWord(Pic::MemoryCode))));
    return false;
  }
  return true;
}

bool Icd2::DebugProgrammer::programAll(const Pic::Memory &mem)
{
  Pic::Memory memory = toDebugMemory(mem, false);
  if ( !programAndVerifyRange(Pic::MemoryCode, memory) ) return false;
  if ( !writeDebugExecutive() ) return false;
  if ( !programAndVerifyRange(Pic::MemoryDebugVector, memory) ) return false;
  if ( !programAndVerifyRange(Pic::MemoryEeprom, memory) ) return false;
  if ( !programAndVerifyRange(Pic::MemoryUserId, memory) ) return false;
  if ( device()->architecture()==Pic::Arch_18X ) {
    if ( !hardware()->command("0C00", 0) ) return false; // #### ??
    QString com = "42" + toHex(0xFB5, 8) + toHex(1, 8); // write RSBUG (?)
    if ( !hardware()->command(com, 0) ) return false;
    if ( !hardware()->command("0C01", 0) ) return false; // #### ??
  }
  if ( !programAndVerifyRange(Pic::MemoryConfig, memory) ) return false;
  return true;
}

bool Icd2::DebugProgrammer::internalRead(Device::Memory *mem, const ::Programmer::VerifyData *vd)
{
  if ( vd==0 || (vd->actions & ::Programmer::BlankCheckVerify) )
    return Icd2::ProgrammerBase::internalRead(mem, 0);
  Pic::Memory memory = toDebugMemory(static_cast<const Pic::Memory &>(vd->memory), true);
  ::Programmer::VerifyData vdata(vd->actions, memory);
  if ( !Icd2::ProgrammerBase::internalRead(0, &vdata) ) return false;
  if ( !readRange(Pic::MemoryDebugVector, 0, &vdata) ) return false;
  return true;
}

bool Icd2::DebugProgrammer::readDebugExecutiveVersion()
{
  if ( !hardware()->getDebugExecVersion(_debugExecutiveVersion) ) return false;
  log(Log::Information, i18n("  Debug executive version: %1").arg(_debugExecutiveVersion.pretty()));
  return true;
}

//----------------------------------------------------------------------------
::Group::SupportType Icd2::DebuggerGroup::supportType(const QString &device) const
{
  if ( !isSupported(device) ) return ::Group::NotSupported;
  return ::Group::SupportType(data(device).debugSupportType);
}

void Icd2::DebuggerGroup::addDevice(const ::Group::DeviceData &data)
{
  if ( FAMILY_DATA[family(data.data->name())].debugExec==0 ) return;
  Group::addDevice(data);
}

::Debugger::Specific *Icd2::DebuggerGroup::createDebuggerSpecific(::Debugger::Base &base) const
{
  const Pic::Data *data = static_cast< ::Debugger::PicBase &>(base).device();
  if ( data==0 ) return 0;
  switch (data->architecture()) {
    case Pic::Arch_16X: return new P16FDebuggerSpecific(base); break;
    case Pic::Arch_17X:
    case Pic::Arch_18X: return new P18FDebuggerSpecific(base); break;
    case Pic::Arch_10X:
    case Pic::Arch_30X:
    case Pic::Nb_Architectures: break;
  }
  Q_ASSERT(false);
  return 0;
}
