/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#ifndef ICD2_DEBUG_H
#define ICD2_DEBUG_H

#include "icd2_prog.h"
#include "devices/pic/pic/pic_debug.h"

namespace Icd2
{
class DebuggerSpecific;

//-----------------------------------------------------------------------------
class DebugProgrammer : public ProgrammerBase
{
Q_OBJECT
public:
  DebugProgrammer(const ::Programmer::Group &group, const Pic::Data *data);
  bool readDebugExecutiveVersion();
  const VersionData &debugExecutiveVersion() const { return _debugExecutiveVersion; }

private:
  VersionData   _debugExecutiveVersion;
  Device::Array _deArray;
  uint          _deStart, _deEnd;

  virtual void clear();
  virtual bool internalSetupHardware();
  virtual bool programAll(const Pic::Memory &memory);
  virtual bool internalRead(Device::Memory *memory, const ::Programmer::VerifyData *vdata);

  bool getDebugExecutive();
  bool writeDebugExecutive();
  Pic::Memory toDebugMemory(const Pic::Memory &memory, bool withDebugExecutive);
};

//-----------------------------------------------------------------------------
class Debugger : public ::Debugger::PicBase
{
public:
  Debugger(DebugProgrammer &programmer) : ::Debugger::PicBase(programmer) {}
  virtual bool setBreakpoints(const QValueList<uint> &addresses);
  Hardware *hardware() { return static_cast<Hardware *>(_programmer.hardware()); }
  DebugProgrammer &programmer() { return static_cast<DebugProgrammer &>(_programmer); }
  DebuggerSpecific *specific();
  bool waitForTargetMode(Pic::TargetMode mode);
  virtual bool readRegister(const Register::TypeData &data, uint &value);
  virtual bool writeRegister(const Register::TypeData &data, uint value);

protected:
  virtual bool internalInit();
  virtual bool internalRun();
  virtual bool internalStep();
  virtual bool softHalt(bool &success);
  virtual bool hardHalt();
  virtual bool internalReset();
  virtual bool updateState();
};

//-----------------------------------------------------------------------------
class DebuggerGroup : public Group
{
public:
  virtual QString name() const { return "icd2_debugger"; }
  virtual QString label() const { return i18n("ICD2 Debugger"); }
  virtual ::Group::SupportType supportType(const QString &device) const;
  virtual ::Programmer::Properties properties() const { return Group::properties() | ::Programmer::Debugger; }
  virtual uint maxNbBreakpoints(const Device::Data *) const { return 1; }

protected:
  virtual void addDevice(const ::Group::DeviceData &data);
  virtual ::Programmer::Base *createBase(const Device::Data *data) const { return new DebugProgrammer(*this, static_cast<const Pic::Data *>(data)); }
  virtual ::Debugger::Base *createDebuggerBase(::Programmer::Base &base) const { return new Debugger(static_cast<DebugProgrammer &>(base)); }
  virtual ::Debugger::Specific *createDebuggerSpecific(::Debugger::Base &base) const;
};

//-----------------------------------------------------------------------------
class DebuggerSpecific : public ::Debugger::Specific
{
public:
  DebuggerSpecific(::Debugger::Base &base) : ::Debugger::Specific(base) {}
  Debugger &base() { return static_cast<Debugger &>(_base); }
  const Debugger &base() const { return static_cast<const Debugger &>(_base); }
  const Pic::Data *device() const { return base().device(); }
  Hardware *hardware() { return base().programmer().hardware(); }
  DebugProgrammer &programmer() { return base().programmer(); }
  ::Debugger::PicSpecific *deviceSpecific() { return base().deviceSpecific(); }
  virtual uint addressWREG() const = 0;
  virtual uint maskPC() const = 0;
  virtual uint addressRegister(uint address) const = 0;
  virtual bool setBreakpoint(uint address) = 0;
  virtual bool readBreakpoint(uint &value) = 0;
  virtual bool init() = 0;
  virtual bool reset() = 0;
};

class P16FDebuggerSpecific : public DebuggerSpecific
{
public:
  P16FDebuggerSpecific(::Debugger::Base &base) : DebuggerSpecific(base) {}
  virtual uint addressBreakpointRegister() const { return 0x18E; }
  virtual uint writeMaskBreakpointRegister() const { return 0x8000; }
  virtual uint readMaskBreakpointRegister() const { return 0x1FFF; }
  virtual uint addressWREG() const { return 0x1BF; }
  virtual uint maskPC() const { return 0x1FFF; }
  virtual uint addressRegister(uint address) const;
  virtual bool setBreakpoint(uint address);
  virtual bool readBreakpoint(uint &value);
  virtual bool init();
  virtual bool reset();
};

class P18FDebuggerSpecific : public DebuggerSpecific
{
public:
  P18FDebuggerSpecific(::Debugger::Base &base) : DebuggerSpecific(base) {}
  virtual uint addressWREG() const;
  virtual uint maskPC() const { return 0xFFFF; }
  virtual uint addressRegister(uint address) const;
  virtual bool setBreakpoint(uint address);
  virtual bool readBreakpoint(uint &value);
  virtual bool init();
  virtual bool reset();

private:
  uint reservedRegisterOffset() const;
};

} // namespace

#endif
