/***************************************************************************
 * Copyright (C) 2006 Nicolas Hadacek <hadacek@kde.org>                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "pickit2.h"

#include "common/global/global.h"

//-----------------------------------------------------------------------------
bool Pickit2::Hardware::readVoltages(VoltagesData &voltages)
{
  log(Log::ExtraDebug, QString("readVoltages: Firmware is %1").arg(_base.firmwareVersion().pretty()));
  if ( _base.firmwareVersion()<VersionData(1, 20, 0) ) {
    log(Log::Warning, i18n("Cannot read voltages with this firmware version."));
    return true;
  }
  if ( !port().command('M') ) return false;
  Pickit::Array a = port().array();
  if ( !port().receive(a) ) return false;
  voltages[Pic::TargetVdd].value = double(a[1] + a[2]*256) * 5.0 / 65535;
  voltages[Pic::TargetVpp].value = double(a[3] + a[4]*256) * (5.0 /65535) * (7.4 / 2.7);
  voltages[Pic::TargetVdd].error = ( a[0] & 0x08 );
  voltages[Pic::TargetVpp].error = ( a[0] & 0x10 );
  return true;
}

bool Pickit2::Hardware::setVddVpp(double vdd, double vpp)
{
  log(Log::ExtraDebug, QString("setVddVpp: Firmware is %1").arg(_base.firmwareVersion().pretty()));
  if ( _base.firmwareVersion()<VersionData(1, 20, 0) ) return true;
  log(Log::NormalDebug, QString("  set Vdd = %1 V and Vpp = %2 V").arg(vdd).arg(vpp));
  Pickit::Array cmd = port().array();
  cmd[0] = 's';
  uint cvdd = uint(32.0 * vdd + 12.5);
  cvdd <<= 6;
  cmd[1] = cvdd & 0xC0;
  cmd[2] = cvdd >> 8;
  uint cvpp = uint(18.61 * vpp);
  cmd[3] = 0x40;
  cmd[4] = cvpp;
  cmd[5] = uchar(4.5 * (cvdd >> 8)); // linit to 75% of vdd
  cmd[6] = uchar(0.75 * cvpp);       // linit to 75% of vpp
  if ( !port().command(cmd) ) return false;
  // wait until vpp is stabilized
  for (uint i=0; i<30; i++) {
    Port::usleep(50000);
    VoltagesData voltages;
    if ( !readVoltages(voltages) ) return false;
    if ( voltages[Pic::TargetVpp].value<(vpp+0.5) ) break;
  }
  return true;
}

void Pickit2::Hardware::command(Pickit::Array &cmd, uchar command, uint nbBytes,
                                uint address, uint i, bool longAddress) const
{
  cmd[i] = command;
  cmd[i+1] = nbBytes;
  cmd[i+2] = address & 0xFF;
  cmd[i+3] = (address >> 8) & 0xFF;
  if (longAddress) cmd[i+4] = (address >> 16) & 0xFF;
}

Pickit::Array Pickit2::Hardware::command(uchar c, uint nbBytes, uint address,
                                         bool longAddress) const
{
  Pickit::Array cmd = port().array();
  command(cmd, c, nbBytes, address, 0, longAddress);
  return cmd;
}

bool Pickit2::Hardware::uploadFirmware(const Pic::Memory &memory)
{
  log(Log::NormalDebug, "  Erase pickit2 device");
  Pickit::Array cmd = command(3, 0xC0, 0x2000); // 0xC0 ??
  if ( !port().command(cmd) ) return false; // erase 0x2000-0x4FFF
  Port::usleep(1000000);
  cmd = command(3, 0xC0, 0x5000); // 0xC0 ??
  if ( !port().command(cmd) ) return false; // erase 0x5000-0x7FFF
  Port::usleep(1000000);

  log(Log::NormalDebug, "  Write firmware");
  Device::Array data = memory.arrayForWriting(Pic::MemoryCode);
  for (uint i=0x1000; i<data.count(); i+=16) { // 2 bytes/word and 32 bytes/command max
    uint start = 0;
    for (; start<16; start++) if ( data[i + start]!=Device::INVALID ) break;
    if ( start==16 ) continue;
    uint end = 15;
    for (; end>start; end--) if ( data[i + end]!=Device::INVALID ) break;
    uint nb = end - start + 1;
    cmd = command(2, 2*nb, 2*i);
    for (uint k=0; k<nb; k++) {
      cmd[5 + 2*k] = data[i + start + k] & 0xFF;
      cmd[6 + 2*k] = (data[i + start + k] >> 8) & 0xFF;
    }
    if ( !port().command(cmd) ) return false;
    Port::usleep(100000);
  }

  log(Log::NormalDebug, "  Verify firmware");
  for (uint i=0x1000; i<data.count(); i+=16) { // 2 bytes/word and 32 bytes/command max
    uint start = 0;
    for (; start<16; start++) if ( data[i + start]!=Device::INVALID ) break;
    if ( start==16 ) continue;
    uint end = 15;
    for (; end>start; end--) if ( data[i + end]!=Device::INVALID ) break;
    uint nb = end - start + 1;
    cmd = command(1, 2*nb, 2*i);
    if ( !port().command(cmd) ) return false;
    QValueVector<uint> read;
    if ( !receiveWords(1, read) ) return false;
    for (uint k=0; k<nb; k++) {
      uint index = i + start + k;
      uint d = read[5 + 2*k] & 0xFF;
      d |= (read[6 + 2*k] << 8);
      uint v = data[index];
      if ( d!=v ) {
        log(Log::Error, i18n("Firmware memory does not match hex file (at address 0x%2: reading 0x%3 and expecting 0x%4).")
                        .arg(QString(toHex(index/2, device().nbCharsAddress())))
                        .arg(QString(toHex(d, device().nbCharsWord(Pic::MemoryCode))))
                        .arg(QString(toHex(v, device().nbCharsWord(Pic::MemoryCode)))));
        return false;
      }
    }
  }

  log(Log::NormalDebug, "  Firmware verified");
  cmd = command(2, 2, 0x7FFE);
  cmd[5] = 0x55;
  cmd[6] = 0x55;
  return port().command(cmd);
}

//-----------------------------------------------------------------------------
bool Pickit2::Baseline::init()
{
  Pickit::Array cmd = hardware().port().array();
  cmd[0] = 'c';
  cmd[1] = 0x00;   // Load Configuration
  cmd[2] = 0x02;   // Load Program latches
  cmd[3] = 0x03;   // Load Data latches
  cmd[4] = 0x04;   // Read Program memory
  cmd[5] = 0x05;   // Read Data latches
  cmd[6] = 0x06;   // Increment Address
  cmd[7] = 0x08;   // Begin programming internally timed
  cmd[8] = 0x08;   // Begin Programming externally timed
  cmd[9] = 0x0E;   // End externally time programming cycle
  cmd[10] = 0x09;  // Bulk Erase program memory
  cmd[11] = 0x0B;  // Bulk Erase Data memory
  cmd[12] = 0xFF;  // Read Program memory
  cmd[13] = 0xFF;  // Read Data latches
  cmd[14] = 0xFF;  // Increment Address
  cmd[15] = 0xFF;  // Begin programming internally timed
  return hardware().port().command(cmd);
}

//-----------------------------------------------------------------------------
bool Pickit2::P16F::init()
{
  Pickit::Array cmd = hardware().port().array();
  cmd[0] = 'c';
  cmd[1] = 0x00;   // Load Configuration
  cmd[2] = 0x02;   // Load Program latches
  cmd[3] = 0x03;   // Load Data latches
  cmd[4] = 0x04;   // Read Program memory
  cmd[5] = 0x05;   // Read Data latches
  cmd[6] = 0x06;   // Increment Address
  cmd[7] = 0x08;   // Begin programming internally timed
  cmd[8] = 0x18;   // Begin Programming externally timed
  cmd[9] = 0x0A;   // End externally time programming cycle
  cmd[10] = 0x09;  // Bulk Erase program memory
  cmd[11] = 0x0B;  // Bulk Erase Data memory
  cmd[12] = 0xFF;  // Read Program memory
  cmd[13] = 0xFF;  // Read Data latches
  cmd[14] = 0xFF;  // Increment Address
  cmd[15] = 0xFF;  // Begin programming internally timed
  return hardware().port().command(cmd);
}

//-----------------------------------------------------------------------------
bool Pickit2::P16F87XA::init()
{
  Pickit::Array cmd = hardware().port().array();
  cmd[0] = 'c';
  cmd[1] = 0x00;   // Load Configuration
  cmd[2] = 0x02;   // Load Program latches
  cmd[3] = 0x03;   // Load Data latches
  cmd[4] = 0x04;   // Read Program memory
  cmd[5] = 0x05;   // Read Data latches
  cmd[6] = 0x06;   // Increment Address
  cmd[7] = 0x08;   // Begin programming internally timed
  cmd[8] = 0x18;   // Begin Programming externally timed
  cmd[9] = 0x17;   // End externally time programming cycle
  cmd[10] = 0x1F;  // Bulk Erase program memory
  cmd[11] = 0x1F;  // Bulk Erase Data memory
  cmd[12] = 0xFF;  // Read Program memory
  cmd[13] = 0xFF;  // Read Data latches
  cmd[14] = 0xFF;  // Increment Address
  cmd[15] = 0xFF;  // Begin programming internally timed
  return hardware().port().command(cmd);
}

//-----------------------------------------------------------------------------
bool Pickit2::P16F7X::init()
{
  Pickit::Array cmd = hardware().port().array();
  cmd[0] = 'c';
  cmd[1] = 0x00;   // Load Configuration
  cmd[2] = 0x02;   // Load Program latches
  cmd[3] = 0x03;   // Load Data latches
  cmd[4] = 0x04;   // Read Program memory
  cmd[5] = 0x05;   // Read Data latches
  cmd[6] = 0x06;   // Increment Address
  cmd[7] = 0x08;   // Begin programming internally timed
  cmd[8] = 0x08;   // Begin Programming externally timed
  cmd[9] = 0x0E;   // End externally time programming cycle
  cmd[10] = 0x09;  // Bulk Erase program memory
  cmd[11] = 0x0B;  // Bulk Erase Data memory
  cmd[12] = 0xFF;  // Read Program memory
  cmd[13] = 0xFF;  // Read Data latches
  cmd[14] = 0xFF;  // Increment Address
  cmd[15] = 0xFF;  // Begin programming internally timed
  return hardware().port().command(cmd);
}

bool Pickit2::P16F716::init()
{
  Pickit::Array cmd = hardware().port().array();
  cmd[0] = 'c';
  cmd[1] = 0x00;   // Load Configuration
  cmd[2] = 0x02;   // Load Program latches
  cmd[3] = 0x03;   // Load Data latches
  cmd[4] = 0x04;   // Read Program memory
  cmd[5] = 0x05;   // Read Data latches
  cmd[6] = 0x06;   // Increment Address
  cmd[7] = 0x08;   // Begin programming internally timed
  cmd[8] = 0x18;   // Begin Programming externally timed
  cmd[9] = 0x0E;   // End externally time programming cycle
  cmd[10] = 0x09;  // Bulk Erase program memory
  cmd[11] = 0x0B;  // Bulk Erase Data memory
  cmd[12] = 0xFF;  // Read Program memory
  cmd[13] = 0xFF;  // Read Data latches
  cmd[14] = 0xFF;  // Increment Address
  cmd[15] = 0xFF;  // Begin programming internally timed
  return hardware().port().command(cmd);
}

//-----------------------------------------------------------------------------
bool Pickit2::P18F::init()
{
  Pickit::Array cmd = hardware().port().array();
  cmd[0] = 'c';
  cmd[1] = 0x00;   // Load Configuration
  cmd[2] = 0x02;   // Load Program latches
  cmd[3] = 0x03;   // Load Data latches
  cmd[4] = 0x04;   // Read Program memory
  cmd[5] = 0x05;   // Read Data latches
  cmd[6] = 0x06;   // Increment Address
  cmd[7] = 0x08;   // Begin programming internally timed
  cmd[8] = 0x18;   // Begin Programming externally timed
  cmd[9] = 0x0A;   // End externally time programming cycle
  cmd[10] = 0x09;  // Bulk Erase program memory
  cmd[11] = 0x0B;  // Bulk Erase Data memory
  cmd[12] = 0xFF;  // Read Program memory
  cmd[13] = 0xFF;  // Read Data latches
  cmd[14] = 0xFF;  // Increment Address
  cmd[15] = 0xFF;  // Begin programming internally timed
  return hardware().port().command(cmd);
}

bool Pickit2::P18F::doEraseCommand(uint cmd1, uint cmd2)
{
  Pickit::Array cmd = hardware().port().array();
  cmd[0] = 0x85;
  cmd[1] = cmd1;
  cmd[2] = cmd2;
  return hardware().port().command(cmd);
}

::Programmer::Result Pickit2::P18F::doEraseCode()
{
  doEraseCommand(0x81, 0x00); // boot
  for (uint i=0; i<device().config().nbBlocks(); i++)
    doEraseCommand(1 << i, 0x80);
  return ::Programmer::Ok;
}

bool Pickit2::P18F::doRead(Pic::MemoryRangeType type, Device::Array &data, const ::Programmer::VerifyData *vdata)
{
  data.resize(device().nbWords(type));
  QValueVector<uint> words;
  switch (type) {
    case Pic::MemoryDeviceId:
    case Pic::MemoryUserId:
    case Pic::MemoryConfig: {
      Pickit::Array cmd = hardware().command(0x80, data.count(), device().range(type).start);
      if ( !hardware().port().command(cmd) ) return false;
      if ( !hardware().receiveWords(1, words) ) return false;
      for (uint k=0; k<data.count(); k++) {
        data[k] = words[k];
        if ( vdata && !hardware().verifyWord(k, data[k], type, *vdata) ) return false;
      }
      break;
    }
    case Pic::MemoryCode:
      for (uint i=0; i<data.count();) {
        Pickit::Array cmd = hardware().port().array();
        for (uint k=0; k<8; k++) hardware().command(cmd, 0x80, 64, 2*(i + 32 * k), 5*k);
        if ( !hardware().port().command(cmd) ) return false;
        for (uint k=0; k<8; k++) {
          if ( !hardware().receiveWords(2, words) ) return false;
          for (uint k=0; k<32; k++) {
            data[i] = words[k];
            if ( vdata && !hardware().verifyWord(i, data[i], type, *vdata) ) return false;
            i++;
          }
        }
      }
      break;
    case Pic::MemoryEeprom:
      for (uint i=0; i<data.count();) {
       Pickit::Array cmd = hardware().command(0x81, 64, i, false);
        if ( !hardware().port().command(cmd) ) return false;
        if ( !hardware().receiveWords(1, words) ) return false;
        for (uint k=0; k<64; k++) {
          data[i] = words[k];
          if ( vdata && !hardware().verifyWord(i, data[i], type, *vdata) ) return false;
          i++;
        }
      }
      break;
    case Pic::MemoryCal:
    case Pic::MemoryCalBackup:
    case Pic::MemoryDebugVector:
    case Pic::MemoryHardwareStack:
    case Pic::MemoryProgramExecutive:
    case Pic::Nb_MemoryRangeTypes: Q_ASSERT(false); return false;
  }
  if ( type==Pic::MemoryCode || type==Pic::MemoryEeprom ) addProgress(data.count());
  return true;
}

bool Pickit2::P18F::doWrite(Pic::MemoryRangeType type, const Device::Array &data, bool force)
{
  Q_UNUSED(force);
  Pickit::Array cmd = hardware().port().array();
  switch (type) {
    case Pic::MemoryUserId: {
      cmd = hardware().command(0x82, data.count() / 2, device().range(type).start);
      for (uint i=0; i<data.count(); i++) cmd[i + 5] = data[i] & 0xFF;
      hardware().port().command(cmd);
      break;
    }
    case Pic::MemoryConfig:
      for (uint i=0; i<data.count()/2; i++) {
        cmd[4*i]     = 0x84;
        cmd[1 + 4*i] = 2*i;
        cmd[2 + 4*i] = data[2*i] & 0xFF;
        cmd[3 + 4*i] = data[2*i+1] & 0xFF;
      }
      hardware().port().command(cmd);
      break;
    case Pic::MemoryCode: {
      uint nb = Pickit2::data(device().name()).progWidth;
      for (uint i=0; i<data.count();) {
        cmd = hardware().command(0x82, nb, 2*i);
        for (uint k=0; k<nb; k++) {
          cmd[5 + 2*k] = data[i] & 0xFF;
          cmd[6 + 2*k] = (data[i] >> 8) & 0xFF;
          i++;
        }
        hardware().port().command(cmd);
      }
      break;
    }
    case Pic::MemoryEeprom:
      for (uint i=0; i<data.count();) {
        cmd = hardware().command(0x87, 32, i, false);
        for (uint k=0; k<32; k++) {
          cmd[4+k] = data[i];
          i++;
        }
        hardware().port().command(cmd);
      }
      break;
    case Pic::MemoryDeviceId:
    case Pic::MemoryCal:
    case Pic::MemoryCalBackup:
    case Pic::MemoryDebugVector:
    case Pic::MemoryHardwareStack:
    case Pic::MemoryProgramExecutive:
    case Pic::Nb_MemoryRangeTypes: Q_ASSERT(false); return false;
  }
  return true;
}

::Programmer::Result Pickit2::P18F::doEraseEeprom()
{
  doEraseCommand(0x84, 0x00);
  return ::Programmer::Ok;
}

bool Pickit2::P18F2X20::doEraseCommand(uint cmd1)
{
  Pickit::Array cmd = hardware().port().array();
  cmd[0] = 0x86;
  cmd[1] = cmd1;
  return hardware().port().command(cmd);
}

::Programmer::Result Pickit2::P18F2X20::doEraseCode()
{
  doEraseCommand(0x83); // boot
  for (uint i=0; i<device().config().nbBlocks(); i++)
    doEraseCommand(0x88 + i);
  return ::Programmer::Ok;
}

::Programmer::Result Pickit2::P18F2X20::doEraseEeprom()
{
  doEraseCommand(0x81);
  return ::Programmer::Ok;
}
