/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *   Copyright (C) 2003-2004 Alain Gibaud <alain.gibaud@free.fr>           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "compile.h"

#include <qtimer.h>
#include <qregexp.h>
#include <klocale.h>

#include "piklab/editor_manager.h"
#include "piklab/main_global.h"
#include "piklab/project.h"
#include "devices/list/device_list.h"
#include "compile_config.h"
#include "common/global/process.h"
#include "common/gui/misc_gui.h"

//-----------------------------------------------------------------------------
const Compile::ArgumentData Compile::ARGUMENT_DATA[] = {
  { "$(SRCPATH)",  I18N_NOOP("Replaced by the source directory.") },
  { "$LKR(xxx)",   I18N_NOOP("Replaced by \"xxx\" when there is a custom linker script.") },
  { "%OBJS",       I18N_NOOP("Replaced by the list of additionnal objects.") },
  { "%LIBS",       I18N_NOOP("Replaced by the list of additionnal libraries.") },
  { "%O",          I18N_NOOP("Replaced by the output filename.") },
  { "%PROJECT",    I18N_NOOP("Replaced by the project name.") },
  { "%COFF",       I18N_NOOP("Replaced by the coff filename.") },
  { "%MAP",        I18N_NOOP("Replaced by the map filename.") },
  { "%SYM",        I18N_NOOP("Replaced by the symbol filename.") },
  { "%LIST",       I18N_NOOP("Replaced by the list filename.") },
  { "%DEVICE",     I18N_NOOP("Replaced by the device name.") },
  { "%FAMILY",     I18N_NOOP("Replaced by the device family name (when needed).") },
  { "%I",          I18N_NOOP("Replaced by the relative input filepath(s).") },
  { "%LKR_PATH",   I18N_NOOP("Replaced by the linker script path.") },
  { "%LKR_NAME",   I18N_NOOP("Replaced by the linker script basename.") },
  { "%LKR",        I18N_NOOP("Replaced by the linker script filename.") },
  { "%SEP",        I18N_NOOP("Replaced by a separation into two arguments.") },
  { 0, 0 }
};

//-----------------------------------------------------------------------------
Compile::LogWidget::LogWidget(QWidget *parent)
  : Log::Widget(parent, "compile_log")
{
  connect(this, SIGNAL(clicked(int, int)), SLOT(lineClicked(int)));
}

void Compile::LogWidget::clear()
{
  Log::Widget::clear();
  _map.clear();
}

void Compile::LogWidget::appendLine(Log::Type type, const QString &message, const PURL::Url &url, uint line)
{
  log(type, message, Log::Delayed);
  if ( !url.isEmpty() ) _map[paragraphs()-1] = Data(url, line);
}

void Compile::LogWidget::appendLine(Log::DebugLevel level, const QString &message, const PURL::Url &url, uint line)
{
  log(level, message, Log::Delayed);
  if ( !url.isEmpty() ) _map[paragraphs()-1] = Data(url, line);
}

void Compile::LogWidget::lineClicked(int line)
{
  if ( !_map.contains(line) ) return;
  Editor *e = Main::editorManager().openEditor(_map[line].url);
  if ( e==0 ) return;
  e->setCursor(_map[line].line, 0);
}

//-----------------------------------------------------------------------------
Compile::Process::Process()
  : _process(0)
{}

void Compile::Process::init(const Tool::Group &group, const Data &data)
{
  _group = &group;
  _config = group.createConfig(const_cast<Project *>(data.project));
  _data = data;
}

Compile::Process::~Process()
{
  delete _config;
}

PURL::Directory Compile::Process::directory(uint i) const
{
  if (_data.project) return _data.project->directory();
  Q_ASSERT( i<_data.items.count() );
  return _data.items[i].url.directory();
}

PURL::Url Compile::Process::url(PURL::FileType type, uint i) const
{
  PURL::Url url;
  if ( _data.project && (_data.category==Tool::Linker || _data.category==Tool::BinToHex) )
    url = _data.project->url().toFileType(PURL::Hex);
  else {
    Q_ASSERT( i<_data.items.count() );
    url = _data.items[i].url;
  }
  if ( type==PURL::Nb_FileTypes ) return url;
  return url.toFileType(type);
}

QString Compile::Process::filepath(PURL::FileType type, uint i) const
{
  return url(type, i).relativeTo(directory(), group().withWine());
}

Compile::FileData Compile::Process::fileData(PURL::FileType type, FileActions actions) const
{
  return FileData(url(type, nbFiles()-1), actions);
}

QStringList Compile::Process::arguments() const
{
  QStringList nargs;
  QStringList args;
  if ( _config->hasCustomArguments(_data.category) ) args = _config->customArguments(_data.category);
  else args = genericArguments(*_config);
  for (uint i=0; i<args.count(); i++) nargs += QStringList::split("%SEP", args[i]);
  args = nargs;
  nargs.clear();
  PURL::Url lkr = group().linkerScript(data().project, data().linkType);
  QRegExp lkrexp("(.*)\\$LKR\\(([^)]*)\\)(.*)");
  for (uint i=0; i<args.count(); i++) {
    if ( args[i].contains("$(SRCPATH)") ) {
      args[i].replace("$(SRCPATH)", directory().path(false));
      args[i].replace("//", "/");
    }
    if ( lkrexp.exactMatch(args[i]) ) {
      args[i] = lkrexp.cap(1);
      if ( hasLinkerScript() ) args[i] += lkrexp.cap(2);
      args[i] += lkrexp.cap(3);
    }
    if ( args[i].contains("%OBJS") ) { // before %O
      if (data().project) nargs += data().project->objectsForLinker(objectExtension(), group().withWine());
      else {
        PURL::Url tmp = url(PURL::Object);
        if ( !objectExtension().isEmpty() ) tmp = tmp.toExtension(objectExtension());
        nargs += tmp.relativeTo(directory(), group().withWine());
      }
      continue;
    }
    if ( args[i].contains("%LIBS") ) {
      if (data().project) nargs += data().project->librariesForLinker(libraryPrefix(), group().withWine());
      continue;
    }
    args[i].replace("%O", outputFilepath());
    args[i].replace("%COFF", filepath(PURL::Coff));
    if ( args[i].contains("%PROJECT") ) {
      if (data().project) args[i].replace("%PROJECT", data().project->name());
      else  args[i].replace("%PROJECT", url().basename());
    }
    args[i].replace("%MAP", filepath(PURL::Map));
    args[i].replace("%SYM", url().toExtension("sym").relativeTo(directory(), group().withWine()));
    args[i].replace("%LIST", filepath(PURL::Lst));
    args[i].replace("%DEVICE", deviceName());
    args[i].replace("%FAMILY", familyName());
    if ( args[i].contains("%I") ) {
      for (uint k=0; k<nbFiles(); k++) nargs += inputFilepath(k);
      continue;
    }
    args[i].replace("%LKR_PATH", lkr.path(false)); // before %LKR
    args[i].replace("%LKR_NAME", lkr.filename()); // before LKR
    args[i].replace("%LKR", lkr.filepath(false));
    if ( !args[i].isEmpty() ) nargs += args[i];
  }
  return nargs;
}

bool Compile::Process::start()
{
  _stdout = QString::null;
  _stderr = QString::null;
  delete _process;
  _process = new ::Process::LineSignal;
  connect(_process, SIGNAL(done(int)), SLOT(done(int)));
  connect(_process, SIGNAL(timeout()), SLOT(timeout()));
  connect(_process, SIGNAL(logStdoutLine(const QString &)), SLOT(logStdoutLine(const QString &)));
  connect(_process, SIGNAL(logStderrLine(const QString &)), SLOT(logStderrLine(const QString &)));
  _process->setWorkingDirectory(directory().path(false));
  QString exec = tool()->executableDirectory().path(false) + tool()->baseExecutable(group().withWine());
  _process->setup(exec, arguments(), group().withWine());
  emit log(Log::Command, _process->arguments().join(" "));
  return _process->start(100000);
}

void Compile::Process::done(int code)
{
  if ( code==0 && successInternal() ) emit success();
  else {
    log(Log::Error, i18n("*** Exited with status: %1 ***").arg(code));
    emit failure();
  }
}

void Compile::Process::timeout()
{
  log(Log::Error, i18n("*** Timeout ***"));
  emit failure();
}

//-----------------------------------------------------------------------------
Compile::Manager::Manager(QObject *parent)
  : QObject(parent, "compile_manager"), _base(0)
{}

void Compile::Manager::cleanFile(const PURL::Url &url)
{
  setupFile(Clean, TodoItem(url, false));
}

void Compile::Manager::compileFile(const TodoItem &item)
{
  _label = i18n("Compiling file...");
  setupFile(Clean | (Main::project() ? CompileOnly : Build), item);
}

void Compile::Manager::setupFile(Operations op, const TodoItem &item)
{
  Log::Base::clear();
  clearAll();
  _operations = op;
  _type = NormalLinking;
  _todo.append(item);
  _action = Compiling;
  QTimer::singleShot(0, this, SLOT(start()));
}

void Compile::Manager::cleanProject(LinkType type)
{
  setupProject(Clean, type);
}

void Compile::Manager::buildProject(LinkType type)
{
  _label = i18n("Building project...");
  setupProject(Clean | Build, type);
}

void Compile::Manager::setupProject(Operations op, LinkType type)
{
  Log::Base::clear();
  clearAll();
  _operations = op;
  _type = type;
  PURL::List files = Main::project()->absoluteFiles();
  PURL::List::const_iterator it = files.begin();
  for (; it!=files.end(); ++it)
    if ( (*it).data().group==PURL::Source ) _todo.append(TodoItem(*it, false));
  _action = Compiling;
  QTimer::singleShot(0, this, SLOT(start()));
}

void Compile::Manager::kill()
{
  if ( clearAll() ) log(Log::Error, i18n("*** Aborted ***"), Log::Delayed);
  emit failure();
}

bool Compile::Manager::clearAll()
{
  bool running = ( _base!=0 );
  delete _base;
  _base = 0;
  _todo.clear();
  return running;
}

void Compile::Manager::clean(const FileData::List &list)
{
  FileData::List::const_iterator it = list.begin();
  for (; it!=list.end(); ++it)
    if ( (*it).actions & Generated ) (*it).url.del(0);
}

bool Compile::Manager::needs(Tool::Category category) const
{
  Tool::ActionType atype = (Main::project() ? Tool::Project : Tool::StandaloneFile);
  return Main::toolGroup().needs(atype, category);
}

void Compile::Manager::setup(Tool::Category category)
{
  delete _base;
  Compile::Data data(category, _items, Main::device(), Main::project(), _type);
  _base = Main::toolGroup().createCompileProcess(data);
}

bool Compile::Manager::setupCompile()
{
  PURL::FileType type = Main::toolGroup().sourceFileType(PURL::Compiler);
  for (uint i=0; i<_items.count(); i++) {
    if ( _items[i].url.fileType()!=type ) {
      if ( _operations!=Clean ) {
        QString e = PURL::extensions(type);
        MessageBox::detailedSorry(i18n("The selected toolchain (%1) cannot compile file. It only supports files with extensions: %2")
                                  .arg(Main::toolGroup().label()).arg(e), i18n("File: %1").arg(_items[i].url.pretty()), Log::Show);
        log(Log::Error, i18n("*** Aborted ***"), Log::Delayed);
        failureSlot();
      }
      return false;
    }
  }
  if ( !compileOnly() && needs(Tool::Assembler) ) {
    PURL::FileType type = Main::toolGroup().sourceFileType(PURL::Assembler);
    for (uint i=0; i<_items.count(); i++)
      _todo.append(TodoItem(_items[i].url.toFileType(type), true));
  }
  setup(Tool::Compiler);
  return true;
}

bool Compile::Manager::setupAssemble()
{
  PURL::FileType type = Main::toolGroup().sourceFileType(PURL::Assembler);
  for (uint i=0; i<_items.count(); i++) {
    if ( _items[i].url.fileType()!=type ) {
      if ( _operations!=Clean ) {
        QString e = PURL::extensions(type);
        MessageBox::detailedSorry(i18n("The selected toolchain (%1) cannot assemble file. It only supports files with extensions: %2")
                                  .arg(Main::toolGroup().label()).arg(e), i18n("File: %1").arg(_items[i].url.pretty()), Log::Show);
        log(Log::Error, i18n("*** Aborted ***"), Log::Delayed);
        failureSlot();
      }
      return false;
    }
  }
  setup(Tool::Assembler);
  return true;
}

bool Compile::Manager::setupLink()
{
  _action = Linking;
  if ( !needs(Tool::Linker) ) {
    start();
    return false;
  }
  setup(Tool::Linker);
  return true;
}

bool Compile::Manager::setupBinToHex()
{
  _action = BinToHex;
  if ( !needs(Tool::BinToHex) ) {
    start();
    return false;
  }
  setup(Tool::BinToHex);
  return true;
}

bool Compile::Manager::prepareAction()
{
  switch (_action) {
    case Compiling:
      if ( _todo.count()!=0 ) {
        _items.clear();
        if ( Main::toolGroup().needAllFiles() ) {
          _items = _todo;
          _todo.clear();
        } else {
          _items.append(_todo[0]);
          _todo.remove(_todo.begin());
        }
        PURL::SourceFamily family = _items[0].url.data().sourceFamily;
        if ( PURL::SOURCE_TYPE_DATA[family].toolType==PURL::Compiler && Main::toolGroup().base(Tool::Compiler) )
          return setupCompile();
        return setupAssemble();
      }
      if ( !compileOnly() ) return setupLink();
      break;
    case Linking: return setupBinToHex();
    case BinToHex: break;
  }
  if ( !(_operations & Clean) ) {
    log(Log::Information, i18n("*** Success ***"), Log::Delayed);
    emit success();
  } else if ( _operations!=Clean ) {
    _operations &= ~Clean;
    if ( Main::project() ) setupProject(_operations, _type);
    else setupFile(_operations, _items[0]);
  }
  return false;
}

void Compile::Manager::start()
{
  delete _base;
  _base = 0;
  if ( !prepareAction() ) return;
  if ( !_base->check() ) {
    failureSlot();
    return;
  }
  if ( _operations & Clean ) {
    clean(_base->files());
    QTimer::singleShot(0, this, SLOT(start()));
    return;
  }
  connect(_base, SIGNAL(success()), SLOT(doneSlot()));
  connect(_base, SIGNAL(failure()), SLOT(failureSlot()));
  connect(_base, SIGNAL(log(Log::Type, const QString &, const PURL::Url &, uint)),
          SLOT(logSlot(Log::Type, const QString &, const PURL::Url &, uint)));
  connect(_base, SIGNAL(log(Log::DebugLevel, const QString &, const PURL::Url &, uint)),
          SLOT(logSlot(Log::DebugLevel, const QString &, const PURL::Url &, uint)));
  if ( !_base->start() ) {
    log(Log::Error, i18n("Failed to execute command: check toolchain configuration"), Log::Delayed);
    failureSlot();
  }
}

void Compile::Manager::logSlot(Log::Type type, const QString &message, const PURL::Url &url, uint line)
{
  if ( type==Log::Error ) setError(message);
  static_cast<LogWidget *>(view())->appendLine(type, message, url, line);
}

void Compile::Manager::logSlot(Log::DebugLevel level, const QString &message, const PURL::Url &url, uint line)
{
  static_cast<LogWidget *>(view())->appendLine(level, message, url, line);
}

void Compile::Manager::doneSlot()
{
  if ( hasError() ) {
    failureSlot();
    return;
  }
  FileData::List list = _base->files();
  FileData::List::const_iterator it = list.begin();
  for (; it!=list.end(); ++it) emit updateFile(*it);
  QTimer::singleShot(0, this, SLOT(start()));
}

void Compile::Manager::failureSlot()
{
  clearAll();
  emit failure();
}
