/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "tool.h"

#include <qdatetime.h>
#include <klocale.h>
#include <kprogress.h>

#include "devices/list/device_list.h"
#include "common/global/process.h"
#include "common/gui/misc_gui.h"
#include "compile_config.h"
#include "compile.h"
#include "tool_config_widget.h"
#include "piklab/project.h"
#include "toolchain_config_widget.h"
#include "piklab/main_global.h"

//----------------------------------------------------------------------------
const Tool::CategoryData Tool::CATEGORY_DATA[Nb_Categories] = {
  { "compiler",     I18N_NOOP("Compiler")      },
  { "assembler",    I18N_NOOP("Assembler")     },
  { "linker",       I18N_NOOP("Linker")        },
  { "bin_to_hex",   I18N_NOOP("Bin to Hex")    },
  { "preprocessor", I18N_NOOP("Preprocessor")  }
};

const Tool::ExecutableTypeData Tool::EXECUTABLE_TYPE_DATA[Nb_ExecutableTypes] = {
  { I18N_NOOP("Unix")    },
  { I18N_NOOP("Windows") }
};

const Compile::DirectoryTypeData Compile::DIRECTORY_TYPE_DATA[Nb_DirectoryTypes] = {
  { "executable",    I18N_NOOP("Executable directory")    },
  { "include",       I18N_NOOP("Header directory")        },
  { "linker_script", I18N_NOOP("Linker Script Directory") },
  { "library",       I18N_NOOP("Library Directory")       },
  { "source",        I18N_NOOP("Source Directory")        }
};

const Compile::FilterData Compile::FILTER_DATA[] = {
  { "warning", I18N_NOOP("warning"), Log::Warning     },
  { "error",   I18N_NOOP("error"),   Log::Error       },
  { "message", I18N_NOOP("message"), Log::Information },
  { 0,         0,                    Log::Nb_Types    }
};

//-----------------------------------------------------------------------------
PURL::Directory Tool::Base::executableDirectory() const
{
  const Compile::Config *config = _group->createConfig(0);
  PURL::Directory dir = config->directory(Compile::ExecutableDirectory);
  delete config;
  return dir;
}

::Process::LineOutput *Tool::Base::checkExecutableProcess(const PURL::Directory &dir, bool withWine) const
{
  ::Process::LineOutput *process = new ::Process::LineOutput;
  process->setup(dir.path(false) + baseExecutable(withWine), checkExecutableOptions(), withWine);
  PURL::Directory wdir = checkExecutableWorkingDirectory();
  if ( !wdir.isEmpty() ) process->setWorkingDirectory(wdir.path(false));
  return process;
}

bool Tool::Base::check(const QString &device) const
{
  bool supported = _group->isSupported(device);
  if ( _group->hasCheckDevicesError() ) {
    MessageBox::sorry(i18n("There were errors detecting supported devices for the selected toolchain (%1). Please check the toolchain configuration.").arg(_group->label()), Log::Show);
    return false;
  }
  if ( !device.isEmpty() && device!=Device::AUTO_DATA.name && !supported ) {
    MessageBox::sorry(i18n("The selected toolchain (%1) does not support device %2.").arg(_group->label()).arg(device), Log::Show);
    return false;
  }
  return true;
}

//-----------------------------------------------------------------------------
Tool::Group::Group()
  : _checkDevicesError(true)
{
  for (uint i=0; i<Nb_Categories; i++) _bases[i] = 0;
}

void Tool::Group::init()
{
  for (uint i=0; i<Nb_Categories; i++) {
    delete _bases[i];
    _bases[i] = baseFactory(Category(i));
    if ( _bases[i]==0 ) continue;
    _bases[i]->_category = Category(i);
    _bases[i]->_group = this;
  }
  ::Group::Base::init();
}

bool Tool::Group::withWine() const
{
  const Compile::Config *config = createConfig(0);
  bool with = config->withWine();
  delete config;
  return with;
}

Compile::Process *Tool::Group::createCompileProcess(const Compile::Data &data) const
{
  Compile::Process *p = processFactory(data.category, data.project);
  Q_ASSERT(p);
  p->init(*this, data);
  return p;
}

ToolConfigWidget *Tool::Group::createConfigWidget(Category category, ::Project *project) const
{
  ToolConfigWidget *cw = configWidgetFactory(category, project);
  if (cw) cw->init(category, *this);
  return cw;
}

Compile::Config *Tool::Group::createConfig(::Project *project) const
{
  Compile::Config *c = configFactory(project);
  c->_group = this;
  return c;
}

ToolchainConfigWidget *Tool::Group::toolchainConfigWidgetFactory(QWidget *parent) const
{
  return new ToolchainConfigWidget(*this, parent);
}

PURL::Directory Tool::Group::autodetectDirectory(Compile::DirectoryType, const PURL::Directory &, bool) const
{
  return PURL::Directory();
}

QString Tool::Group::defaultLinkerScriptFilename(Compile::LinkType type, const QString &device) const
{
  QString basename = device.lower();
  if ( type==Compile::Icd2Linking ) basename += 'i';
  return basename + '.' + PURL::extension(PURL::Lkr);
}

bool Tool::Group::hasCustomLinkerScript(const ::Project *project) const
{
  return ( project && !project->customLinkerScript().isEmpty() );
}

PURL::Url Tool::Group::linkerScript(const ::Project *project, Compile::LinkType type) const
{
  if ( hasCustomLinkerScript(project) ) return project->customLinkerScript();
  const Compile::Config *config = createConfig(const_cast< ::Project *>(project));
  QString filename = defaultLinkerScriptFilename(type, config->device());
  PURL::Url url = PURL::Url(config->directory(Compile::LinkerScriptDirectory), filename);
  delete config;
  return url;
}

::Process::LineOutput *Tool::Group::checkDevicesProcess(uint i, const PURL::Directory &dir, bool withWine) const
{
  ::Process::LineOutput *process = new ::Process::LineOutput;
  Tool::Category cat = checkDevicesCategory();
  process->setup(dir.path(false) + base(cat)->baseExecutable(withWine), checkDevicesOptions(i), withWine);
  return process;
}

void Tool::Group::initSupported()
{
  _checkDevicesError = false;
  Tool::Category cat = checkDevicesCategory();
  QValueList<const Device::Data *> list;
  if ( cat==Tool::Nb_Categories ) list = getSupportedDevices(QString::null);
  else {
    const Compile::Config *config = createConfig(0);
    PURL::Directory dir = config->directory(Compile::ExecutableDirectory);
    delete config;
    for (uint i=0; i<nbCheckDevices(); i++) {
      QStringList lines;
      ::Process::LineOutput *process = checkDevicesProcess(i, dir, withWine());
      ::Process::State state = ::Process::runSynchronously(*process, ::Process::Start, 10000);
      if ( state==::Process::Exited ) {
        QStringList lines = process->sout() + process->serr();
        list += getSupportedDevices(lines.join("\n"));
      } else _checkDevicesError = true;
      delete process;
    }
  }
  QValueList<const Device::Data *>::const_iterator it;
  for (it=list.begin(); it!=list.end(); ++it) {
    ::Group::DeviceData data;
    data.data = *it;
    data.supportType = ::Group::Tested;
    addDevice(data);
  }
}
