/***************************************************************************
 *   Copyright (C) 2005-2006 Nicolas Hadacek <hadacek@kde.org>             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#ifndef TOOL_H
#define TOOL_H

#include <qstringlist.h>
#include <qmap.h>
#include <klocale.h>

#include "common/global/purl.h"
#include "common/common/group.h"
class Project;
class ToolConfigWidget;
class ToolchainConfigWidget;
namespace Device { class Data; class Memory; }
namespace Process { class LineOutput; }

namespace Tool
{
//----------------------------------------------------------------------------
  enum Category { Compiler = 0, Assembler, Linker, BinToHex, Preprocessor, Nb_Categories };
  struct CategoryData {
    const char *name, *label;
  };
  extern const CategoryData CATEGORY_DATA[Nb_Categories];

  enum ExecutableType { Unix = 0, Windows, Nb_ExecutableTypes };
  struct ExecutableTypeData {
    const char *label;
  };
  extern const ExecutableTypeData EXECUTABLE_TYPE_DATA[Nb_ExecutableTypes];
} // namespace

namespace Compile
{
  enum LinkType { NormalLinking, Icd2Linking };
  class Process;
  class Config;

  class TodoItem {
  public:
    TodoItem() {}
    TodoItem(const PURL::Url &_url, bool _generated) : url(_url), generated(_generated) {}
    PURL::Url url;
    bool generated;
  };

  class Data {
  public:
    Data() {}
    Data(Tool::Category c, const QValueList<TodoItem> &i, const QString &d, const Project *p, LinkType lt)
      : category(c), items(i), device(d), project(p), linkType(lt) {}
    Tool::Category category;
    QValueList<TodoItem> items;
    QString  device;
    const Project *project;
    LinkType linkType;
  };

  enum DirectoryType { ExecutableDirectory = 0, HeaderDirectory, LinkerScriptDirectory, LibraryDirectory, SourceDirectory, Nb_DirectoryTypes };
  struct DirectoryTypeData {
    const char *key, *label;
  };
  extern const DirectoryTypeData DIRECTORY_TYPE_DATA[Nb_DirectoryTypes];

} // namespace

namespace Tool
{
  enum ActionType { StandaloneFile, Project };
  class Group;

//-----------------------------------------------------------------------------
class Base
{
public:
  virtual ~Base() {}
  virtual QString baseExecutable(bool withWine) const = 0;
  PURL::Directory executableDirectory() const;
  virtual QStringList checkExecutableOptions() const = 0;
  virtual PURL::Directory checkExecutableWorkingDirectory() const { return PURL::Directory(); }
  ::Process::LineOutput *checkExecutableProcess(const PURL::Directory &dir, bool withWine) const;
  virtual bool checkExecutableResult(bool withWine, QStringList &lines) const = 0;
  bool check(const QString &device = QString::null) const;

protected:
  Category     _category;
  const Group *_group;

  friend class Group;
};

//-----------------------------------------------------------------------------
class Group : public ::Group::Base
{
public:
  Group();
  virtual void init();
  const ::Tool::Base *base(Category category) const { return _bases[category]; }
  virtual QString downloadText() const = 0;
  virtual ExecutableType preferedExecutableType() const = 0;
  virtual bool hasDirectory(Compile::DirectoryType) const { return false; }
  virtual PURL::FileType linkerScriptType() const { return PURL::Nb_FileTypes; }
  virtual PURL::Directory autodetectDirectory(Compile::DirectoryType type, const PURL::Directory &execDir, bool withWine) const;
  bool withWine() const;
  virtual uint nbCheckDevices() const { return 1; }
  bool hasCheckDevicesError() const { return _checkDevicesError; }
  virtual Tool::Category checkDevicesCategory() const = 0;
  virtual QStringList checkDevicesOptions(uint) const { return QStringList(); }
  ::Process::LineOutput *checkDevicesProcess(uint i, const PURL::Directory &execDir, bool withWine) const;
  virtual QValueList<const Device::Data *> getSupportedDevices(const QString &s) const = 0;
  virtual bool needAllFiles() const { return false; }
  virtual bool needs(ActionType type, Category category) const = 0;
  Compile::Process *createCompileProcess(const Compile::Data &data) const;
  ToolConfigWidget *createConfigWidget(Category category, ::Project *project) const;
  Compile::Config *createConfig(::Project *project) const;
  bool hasCustomLinkerScript(const ::Project *project) const;
  virtual PURL::Url linkerScript(const ::Project *project, Compile::LinkType type) const;
  virtual PURL::FileType sourceFileType(PURL::ToolType type) const = 0;
  virtual QString generateDeviceConfig(PURL::ToolType type, const Device::Memory &memory, bool &ok) const = 0;
  virtual Compile::Process *processFactory(Category category, const ::Project *project) const = 0;
  virtual ToolConfigWidget *configWidgetFactory(Category category, ::Project *project) const = 0;
  virtual ToolchainConfigWidget *toolchainConfigWidgetFactory(QWidget *parent) const;

protected:
  virtual void initSupported();
  virtual Tool::Base *baseFactory(Category category) const = 0;
  virtual QString defaultLinkerScriptFilename(Compile::LinkType type, const QString &device) const;
  virtual Compile::Config  *configFactory(::Project *project) const = 0;

private:
  ::Tool::Base *_bases[Nb_Categories];
  bool _checkDevicesError;
};

} // namespace

#endif
