/***************************************************************************
 *   Copyright (C) 2006 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "ccsc_compile.h"

#include <qregexp.h>
#include <klocale.h>

#include "ccsc_config.h"
#include "piklab/project.h"
#include "devices/list/device_list.h"
#include "devices/pic/base/pic.h"

const Compile::FilterData CCSC::FILTER_DATA[] = {
  { ">>>", I18N_NOOP("warning"), Log::Warning     },
  { "***", I18N_NOOP("error"),   Log::Error       },
  { "---", I18N_NOOP("message"), Log::Information },
  { 0,     0,                    Log::Nb_Types    }
};

QString CCSC::CompileFile::familyName() const
{
  const Pic::Data *pdata = static_cast<const Pic::Data *>(Device::lister().data(data().device));
  switch (pdata->architecture()) {
    case Pic::Arch_10X: return "B";
    case Pic::Arch_16X: return "M";
    case Pic::Arch_17X: break;
    case Pic::Arch_18X: return "H";
    case Pic::Arch_30X: break;
    case Pic::Nb_Architectures: break;
  }
  Q_ASSERT(false);
  return QString::null;
}

QStringList CCSC::CompileFile::genericArguments(const Compile::Config &config) const
{
  QStringList args;
  args += "+STDERR"; // undocumented
  args += "+EA";     // show all messages and warnings
  args += "-P";      // close compile windows after compilation done
  args += "+DF";     // output COFF file
  args += "+LSlst";  // normal list file
  args += "+O8hex";  // produce 8bit Intel file
  args += "+M";      // generate symbol file
  args += "-J";      // do not create project file
  args += "-A";      // do not create stat file
  args += "+F%FAMILY";
  args += config.includeDirs(Tool::Compiler, "I+=\"", "\"");
  args += config.customOptions(Tool::Compiler);
  args += "%I";
  return args;
}

void CCSC::CompileFile::logStderrLine(const QString &line)
{
  uint i = 0;
  for (; FILTER_DATA[i].name; i++) if ( line.startsWith(FILTER_DATA[i].name) ) break;
  if ( FILTER_DATA[i].name==0 ) {
    emit log(Log::Data, line);
    return;
  }
  QRegExp rexp("[*>-]+\\s\\w+\\s\\d+\\s\"([^\"]*)\"\\sLine\\s(\\d+)\\([^)]*\\):(.*)");
  if ( rexp.exactMatch(line) ) {
    PURL::Url url(rexp.cap(1));
    emit log(FILTER_DATA[i].type, url.filename() + ":" + rexp.cap(2) + ": "
             + i18n(FILTER_DATA[i].label) + ": " + rexp.cap(3).stripWhiteSpace(), url, rexp.cap(2).toUInt()-1);
    return;
  }
  rexp.setPattern("[*>-]+\\s\"([^\"]*)\"\\sLine\\s(\\d+):\\s\\w+\\s#\\d+:(.*)");
  if ( rexp.exactMatch(line) ) {
    PURL::Url url(rexp.cap(1));
    emit log(FILTER_DATA[i].type, url.filename() + ":" + rexp.cap(2) + ": "
             + i18n(FILTER_DATA[i].label) + ": " + rexp.cap(3).stripWhiteSpace(), url, rexp.cap(2).toUInt()-1);
    return;
  }
  emit log(FILTER_DATA[i].type, line);
}

Compile::FileData::List CCSC::CompileFile::files() const
{
  Compile::FileData::List files;
  files += fileData(PURL::Lst, Compile::Generated | Compile::InProject);
  files += fileData(PURL::Hex, Compile::Generated | Compile::InProject);
  files += fileData(PURL::Coff, Compile::Generated | Compile::InProject);
  files += Compile::FileData(url().toExtension("sym"), Compile::Generated);
  files += Compile::FileData(url().toExtension("err"), Compile::Generated);
  files += Compile::FileData(url().toExtension("esym"), Compile::Generated);
  return files;
}
