/***************************************************************************
 *   Copyright (C) 2006 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "sdcc.h"

#include "devices/list/device_list.h"
#include "sdcc_compile.h"
#include "sdcc_config.h"
#include "devices/pic/pic/pic_memory.h"
#include "common/global/pfile.h"
#include "common/global/process.h"
#include "tools/base/tool_list.h"

//----------------------------------------------------------------------------
const SDCC::FamilyData SDCC::FAMILY_DATA[Nb_Families] = {
  { "pic14" },
  { "pic16" }
};

QString SDCC::toDeviceName(const QString &device)
{
  if ( device.startsWith("PS") ) return device.lower();
  return "p" + device.lower();
}

SDCC::Family SDCC::family(const QString &device)
{
  const Device::Data *data = Device::lister().data(device);
  switch (static_cast<const Pic::Data *>(data)->architecture()) {
    case Pic::Arch_10X:
    case Pic::Arch_16X: return P14;
    case Pic::Arch_17X:
    case Pic::Arch_18X: return P16;
    case Pic::Arch_30X:
    case Pic::Nb_Architectures: break;
  }
  Q_ASSERT(false);
  return Nb_Families;
}

//----------------------------------------------------------------------------
bool SDCC::Base::checkExecutableResult(bool, QStringList &lines) const
{
  return ( lines.count()>0 && lines[0].startsWith("SDCC") );
}

//----------------------------------------------------------------------------
QString SDCC::Group::downloadText() const
{
  return i18n("The Small Devices C Compiler can be downloaded from <a href=\"http://sdcc.sourceforge.net\">this page</a>.<br>"
              "A binary package for Mandriva is available <a href=\"http://sourceforge.net/project/showfiles.php?group_id=138852&package_id=181223\">here</a>.");
}

Tool::Base *SDCC::Group::baseFactory(Tool::Category category) const
{
  if ( category==Tool::Compiler || category==Tool::Preprocessor || category==Tool::Linker) return new ::SDCC::Base;
  return 0;
}

SDCC::Group::Group()
  : _checkDevicesTmp(0, ".c")
{
  // used to obtain device list
  if ( _checkDevicesTmp.openForWrite() ) _checkDevicesTmp.appendText("void main(void) {}\n");
  _checkDevicesTmp.close();
}

QStringList SDCC::Group::checkDevicesOptions(uint i) const
{
  QStringList options;
  options += QString("-m") + SDCC::FAMILY_DATA[i].name;
  options += "-phelp";
  options += _checkDevicesTmp.url().filepath(false);
  return options;
}

QValueList<const Device::Data *> SDCC::Group::getSupportedDevices(const QString &s) const
{
  QValueList<const Device::Data *> list;
  QStringList devices = QStringList::split(' ', s.simplifyWhiteSpace().upper());
  for (uint i=0; i<devices.count(); i++) {
    QString name = devices[i];
    if ( devices[i].startsWith("P1") ) name = name.mid(1);
    const Device::Data *data = Device::lister().data(name);
    if (data) list.append(data);
  }
  return list;
}

bool SDCC::Group::needs(Tool::ActionType type, Tool::Category category) const
{
  if ( type==Tool::Project && category==Tool::Linker ) return true;
  return false;
}

Compile::Process *SDCC::Group::processFactory(Tool::Category category, const Project *project) const
{
  switch (category) {
    case Tool::Compiler:
      if (project) return new SDCC::CompileProjectFile;
      return new SDCC::CompileStandaloneFile;
    case Tool::Linker:
      return new SDCC::LinkProjectFile;
    default: break;
  }
  return 0;
}

ToolConfigWidget *SDCC::Group::configWidgetFactory(Tool::Category category, Project *project) const
{
  switch (category) {
    case Tool::Compiler: return new SDCC::CompilerConfigWidget(project);
    case Tool::Linker: return new SDCC::LinkerConfigWidget(project);
    default: break;
  }
  return 0;
}

Compile::Config *SDCC::Group::configFactory(::Project *project) const
{
  return new Config(project);
}

PURL::FileType SDCC::Group::sourceFileType(PURL::ToolType type) const
{
  if ( type==PURL::Assembler ) return PURL::AsmGPAsm;
  if ( type==PURL::Compiler ) return PURL::CSource;
  return PURL::Nb_FileTypes;
}

QString SDCC::Group::generateDeviceConfig(PURL::ToolType type, const Device::Memory &memory, bool &ok) const
{
  if ( type==PURL::Assembler ) return Tool::lister().group("gputils")->generateDeviceConfig(type, memory, ok);
  const Pic::Memory &pmemory = static_cast<const Pic::Memory &>(memory);
  const Pic::Data &data = pmemory.device();
  const Pic::Config &config = data.config();
  ok = true;
  QString s = "typedef unsigned int word;\n";
  uint address = data.range(Pic::MemoryConfig).start;
  for (uint i=0; i<data.nbWords(Pic::MemoryConfig); i++) {
    uint v = pmemory.normalizedWord(Pic::MemoryConfig, i);
    const Pic::Config::Word &cword = config._words[i];
    if ( cword.masks.count()==0 ) continue;
    s += "word at ";
    QString cn = config.configName(i, Pic::Config::ConfigNameCMacro);
    if ( cn.isEmpty() ) s += toHexLabel(address, data.nbCharsAddress());
    else s += cn;
    s += " " + config.configName(i, Pic::Config::ConfigNameCVariable) + " = ";
    QStringList cnames;
    for (uint k=0; k<cword.masks.count(); k++) {
      const Pic::Config::Mask &cmask = cword.masks[k];
      if ( cmask.name=="DEBUG" ) continue;
      if ( (cmask.value & cword.pmask)==cmask.value ) continue; // protected bits
      for (int l=cmask.values.count()-1; l>=0; l--) {
        if ( (v & cmask.values[l].value)==cmask.values[l].value ) {
          if ( cmask.values[l].cname.isEmpty() ) ok = false;
          else cnames += cmask.values[l].cname;
          break;
        }
      }
    }
    s += cnames.join(" & ");
    s += ";";
    s += "\n";
    address += data.addressIncrement(Pic::MemoryConfig);
  }
  return s;
}
